# (c) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from elasticsearch import exceptions as es_exc
from searchlight.common import utils as common_utils
from searchlight.elasticsearch.plugins import utils as plugin_utils
from searchlight.tests import utils as test_utils


class TestReindexingUtils(test_utils.BaseTestCase):
    def test_reindex(self):
        src = 'sl-old'
        dst = 'sl-new'
        single = ['OS::Neutron::Net']
        mult = ['OS::Neutron::Net', 'OS::Nova::Server', 'OS::Glance::Image']
        reindex_name = 'searchlight.elasticsearch.plugins.utils.helper_reindex'

        expected_single = {'query': {
                           'bool': {
                               'filter': {
                                   'terms': {
                                       '_type': single}
                               }
                           }}, 'version': 'true'}
        expected_mult = {'query': {
                         'bool': {
                             'filter': {
                                 'terms': {
                                     '_type': mult}
                             }
                         }}, 'version': 'true'}

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            with mock.patch(reindex_name) as mock_reindex:
                # Plug in the ES mock.
                mock_api.return_value = mock_engine

                # Test #1: Reindex a single index.
                plugin_utils.reindex(src, dst, single)
                mock_api.assert_called_with()
                mock_reindex.assert_called_with(client=mock_engine,
                                                source_index=src,
                                                target_index=dst,
                                                query=expected_single)

                # Test #2: Reindex multiple indexes.
                plugin_utils.reindex(src, dst, mult)
                mock_reindex.assert_called_with(client=mock_engine,
                                                source_index=src,
                                                target_index=dst,
                                                query=expected_mult)

    @mock.patch('searchlight.elasticsearch.plugins.'
                'utils._get_index_settings_from_config', return_value={})
    def test_create_new_index(self, mock_get_settings):
        # Regex for matching the index name. The index name is the group
        # group name appended with a time stmap. The format for the
        # timestamp is defined in elasitcsearch.plugins.utils and is
        # defined as:
        #     [4 digit Year] [2 digit Month] [2 digit Day] [2 digit Hour]
        #     [2 digit Minutes] [2 digit Seconds]
        # We want to search for this pattern exactly, which is why we are
        # specifying "^" and "$" in the Regex. We elected to make the unit
        # test more complicated, rather than artificially wrap datetime
        # functionality in the code just for the tests.
        TS_FORMAT = '\d{4}_\d{2}_\d{2}_\d{2}_\d{2}_\d{2}$'
        group = 'searchlight'

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Test #1: Create a new index.
            index_name = plugin_utils.create_new_index(group)

            self.assertRegex(index_name, group + '-' + TS_FORMAT)
            mock_api.assert_called_with()
            mock_engine.indices.create.assert_called_with(index=index_name)

    def test_add_extra_mappings(self):
        # Set up the ES mock.
        mock_engine = mock.Mock()
        index = 'fake'
        b_parent = {'dynamic': 'strict',
                    '_parent': {
                        'type': 'never_used_parent'},
                    'properties': {
                    }}
        b_orphan = {'dynamic': 'strict',
                    'properties': {
                    }}
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Test #1: One doc_type that already exists. With parent.
            mock_engine.indices.exists_type.return_value = True
            docs = [('doc1', True)]

            plugin_utils.add_extra_mappings(index_name=index,
                                            doc_type_info=docs)
            mock_engine.indices.put_mapping.assert_not_called()

            # Test #2: One doc_type that already exists. Without parent.
            mock_engine.indices.exists_type.return_value = True
            docs = [('doc1', False)]

            plugin_utils.add_extra_mappings(index_name=index,
                                            doc_type_info=docs)
            mock_engine.indices.put_mapping.assert_not_called()

            # Test #3: One doc_type that does not exist. With parent.
            mock_engine.indices.exists_type.return_value = False
            docs = [('doc1', True)]

            plugin_utils.add_extra_mappings(index_name=index,
                                            doc_type_info=docs)
            mock_engine.indices.put_mapping.assert_called_with(index=index,
                                                               doc_type='doc1',
                                                               body=b_parent)

            # Test #4: One doc_type that does not exist. Without parent.
            mock_engine.indices.exists_type.return_value = False
            mock_engine.indices.put_mapping.reset_mock()
            docs = [('doc1', False)]

            plugin_utils.add_extra_mappings(index_name=index,
                                            doc_type_info=docs)
            mock_engine.indices.put_mapping.assert_called_with(index=index,
                                                               doc_type='doc1',
                                                               body=b_orphan)

            # Test #5: Two doc_type's that do not exist. Mixed parents.
            mock_engine.indices.exists_type.return_value = False
            mock_engine.indices.put_mapping.reset_mock()
            docs = [('doc1', True), ('doc2', False)]
            calls = [mock.call(index=index, doc_type='doc1', body=b_parent),
                     mock.call(index=index, doc_type='doc2', body=b_orphan)]

            plugin_utils.add_extra_mappings(index_name=index,
                                            doc_type_info=docs)
            mock_engine.indices.put_mapping.assert_has_calls(calls)

    def test_setup_alias(self):
        ndx = 'sl'
        ndx_s = 'sl-s'
        ndx_l = 'sl-l'
        body = {'actions': [{'add': {'index': ndx, 'alias': ndx_l}}]}

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Test #1: No search alias, no listener alias.
            mock_engine.indices.exists_alias.side_effect = [False, False]
            plugin_utils.setup_alias(ndx, ndx_s, ndx_l)
            calls = [mock.call(index=ndx, name=ndx_s),
                     mock.call(index=ndx, name=ndx_l)]
            mock_engine.indices.put_alias.assert_has_calls(calls)
            mock_engine.indices.update_aliases.assert_not_called()
            mock_api.assert_called_with()

            # Test #2: No search alias, existing listener alias.
            mock_engine.reset_mock()
            mock_engine.indices.exists_alias.side_effect = [False, True]
            plugin_utils.setup_alias(ndx, ndx_s, ndx_l)
            mock_engine.indices.put_alias.assert_called_once_with(
                index=ndx, name=ndx_s)
            mock_engine.indices.update_aliases.assert_called_with(body=body)

            # Test #3: Existing search alias, existing listener alias.
            mock_engine.reset_mock()
            mock_engine.indices.exists_alias.side_effect = [True, True]
            plugin_utils.setup_alias(ndx, ndx_s, ndx_l)
            mock_engine.indices.put_aliases.assert_not_called()
            mock_engine.indices.update_aliases.assert_called_with(body=body)

            # Test #4: Exception while creating search alias
            mock_engine.reset_mock()
            mock_engine.indices.exists_alias.side_effect = [False]
            mock_engine.indices.put_alias.side_effect = [TypeError]
            self.assertRaises(TypeError, plugin_utils.setup_alias,
                              index_name=ndx, alias_search=ndx_s,
                              alias_listener=ndx_l)
            mock_engine.indices.delete.assert_called_once_with(index=ndx)

            # Test #5: Exception while creating listener alias
            mock_engine.reset_mock()
            mock_engine.indices.exists_alias.side_effect = [True, False]
            mock_engine.indices.put_alias.side_effect = [TypeError]
            self.assertRaises(TypeError, plugin_utils.setup_alias,
                              index_name=ndx, alias_search=ndx_s,
                              alias_listener=ndx_l)
            mock_engine.indices.delete.assert_called_once_with(index=ndx)

    def test_alias_search_update(self):
        ndx = 'sl-search'
        old_ndx = 'sl-old'
        new_ndx = 'sl-new'
        alias = {old_ndx: {'aliases': {ndx: {}}}}

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Test #1: Existing search alias, different index.
            mock_engine.indices.get_alias.return_value = alias

            index = plugin_utils.alias_search_update(ndx, new_ndx)
            body = {
                'actions': [{'remove': {'index': old_ndx, 'alias': ndx}},
                            {'add': {'index': new_ndx, 'alias': ndx}}]}
            mock_engine.indices.update_aliases.assert_called_once_with(body)
            self.assertEqual(index, old_ndx)
            mock_api.assert_called_with()

            # Test #2: Existing search alias, indexes are the same.
            mock_engine.reset_mock()
            mock_engine.indices.get_alias.return_value = alias

            index = plugin_utils.alias_search_update(ndx, old_ndx)
            mock_engine.indices.update_aliases.assert_not_called()
            self.assertIsNone(index)

            # Test #3: No index.
            mock_engine.reset_mock()

            index = plugin_utils.alias_search_update(ndx, None)
            self.assertIsNone(index)
            mock_engine.indices.get_alias.assert_not_called()
            mock_engine.indices.update_aliases.assert_not_called()

            # Test #4: Alias update failure.
            mock_engine.reset_mock()
            mock_engine.indices.get_alias.return_value = alias
            mock_engine.indices.update_aliases.side_effect = [TypeError]

            self.assertRaises(TypeError, plugin_utils.alias_search_update,
                              alias_search=ndx, index_name=new_ndx)

            # Test #5: No search alias.
            mock_engine.reset_mock()
            mock_engine.indices.get_alias.side_effect = es_exc.NotFoundError
            mock_engine.indices.update_aliases.side_effect = [None]

            index = plugin_utils.alias_search_update(ndx, new_ndx)
            body = {'actions': [{'add': {'index': new_ndx, 'alias': ndx}}]}
            mock_engine.indices.update_aliases.assert_called_once_with(body)
            self.assertIsNone(index)

    def test_delete_index(self):
        old_ndx = 'sl-old'

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Delete the existing index.
            plugin_utils.delete_index(old_ndx)

            mock_engine.indices.delete.assert_called_once_with(ignore=404,
                                                               index=old_ndx)
            mock_api.assert_called_with()

    def test_delete_non_exist_index(self):
        old_ndx = None

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Delete the existing index.
            plugin_utils.delete_index(old_ndx)

            mock_engine.indices.delete.assert_not_called()
            mock_api.assert_not_called()

    def test_alias_error_cleanup(self):
        single = {'g': 'ndx'}
        multiple = {'g1': 'ndx1', 'g2': 'ndx2', 'g3': 'ndx3'}

        # Set up the ES mock.
        mock_engine = mock.Mock()
        with mock.patch('searchlight.elasticsearch.get_api') as mock_api:
            # Plug in the ES mock.
            mock_api.return_value = mock_engine

            # Test #1: Cleanup a single index.
            plugin_utils.alias_error_cleanup(single)
            mock_engine.indices.delete.assert_called_once_with(ignore=404,
                                                               index='ndx')
            mock_api.assert_called_with()

            # Test #2: Cleanup multiple indexes.
            mock_engine.reset_mock()
            plugin_utils.alias_error_cleanup(multiple)
            calls = [mock.call(ignore=404, index='ndx1'),
                     mock.call(ignore=404, index='ndx2'),
                     mock.call(ignore=404, index='ndx3')]
            mock_engine.indices.delete.assert_has_calls(calls, any_order=True)

            # Test #3: Cleanup multiple indexes with an error.
            mock_engine.reset_mock()
            mock_engine.indices.delete.side_effect = [None, Exception, None]

            plugin_utils.alias_error_cleanup(multiple)
            calls = [mock.call(ignore=404, index='ndx1'),
                     mock.call(ignore=404, index='ndx2'),
                     mock.call(ignore=404, index='ndx3')]
            mock_engine.indices.delete.assert_has_calls(calls, any_order=True)

    def test_type_expansion(self):
        doc_types = ['simple-plugin', 'simple-child', 'something-else']

        # Don't do any expansion
        self.assertEqual(doc_types,
                         common_utils.expand_type_matches(doc_types,
                                                          doc_types))

        self.assertEqual([], common_utils.expand_type_matches([], doc_types))

        # Expansion time!
        self.assertEqual(
            ['simple-plugin', 'simple-child'],
            common_utils.expand_type_matches(['simple-*'], doc_types))

        self.assertEqual(
            ['something-else'],
            common_utils.expand_type_matches(['something-*'], doc_types))

        # Doesn't match. Sad!
        self.assertEqual(
            ['does-not-match-*'],
            common_utils.expand_type_matches(['does-not-match-*'], doc_types))

        self.assertEqual(
            doc_types,
            common_utils.expand_type_matches(['simple-*', 'something-*'],
                                             doc_types))



# Copyright 2015 Hewlett-Packard Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import mock

from searchlight.elasticsearch.plugins import designate
from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight import pipeline
from searchlight.tests.unit.test_designate_recordset_plugin import \
    _recordset_fixture
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


now = datetime.datetime.utcnow()
five_mins = datetime.timedelta(minutes=5)
updated_now = now.strftime('%Y-%m-%dT%H:%M:%SZ')
created_now = (now - five_mins).strftime('%Y-%m-%dT%H:%M:%SZ')

USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'

TENANT1 = u'4d64ac83-87af-4d2a-b884-cc42c3e8f2c0'
TENANT2 = u'c6993374-7c4b-4f18-b317-85e3acdfd259'

# Zone ids
ID1 = u'9b4c2de5-dead-0000-beef-e3b99695aa69'
ID2 = u'0c3dae67-c4d3-42d6-9d8d-98da56321bca'
ID3 = u'7c3df12a-42d6-9d8d-c4d3-98da56321999'


def _zone_fixture(zone_id, tenant_id, name, **kwargs):
    zone = {
        'tenant_id': tenant_id,
        'name': name,
        'status': 'pending',
        'created_at': created_now,
        'updated_at': updated_now
    }
    zone.update(**kwargs)
    return zone


class TestZonePlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestZonePlugin, self).setUp()
        self.plugin = designate.zones.ZoneIndex()
        self.plugin.child_plugins = [designate.recordsets.RecordSetIndex()]

        self._create_fixtures()

        self.mock_session = mock.Mock()
        self.mock_session.get_endpoint.return_value = \
            'http://localhost/glance/v2'
        patched_ses = mock.patch(
            'searchlight.elasticsearch.plugins.openstack_clients._get_session',
            return_value=self.mock_session)
        patched_ses.start()
        self.addCleanup(patched_ses.stop)

    def _create_fixtures(self):
        self.zone1 = _zone_fixture(ID1, TENANT1, name='test.com.',
                                   type='PRIMARY')
        self.zone2 = _zone_fixture(ID2, TENANT1, name='other.com.',
                                   type='PRIMARY')
        self.zone3 = _zone_fixture(ID3, TENANT2, name='tenant2.com.',
                                   type='PRIMARY')
        self.zones = (self.zone1, self.zone2, self.zone3)

    def check_item(
            self, item, expected_type, event_type, payload, plugin_type):
        self.assertIsInstance(item, expected_type)
        self.assertEqual(event_type, item.event_type)
        self.assertIsInstance(item.plugin, plugin_type)
        self.assertEqual(payload, item.payload)

    def test_missing_updated(self):
        """Designate records don't always have a value for 'updated'"""
        zone_to_test = dict(self.zone1)
        zone_to_test['updated_at'] = None
        serialized = self.plugin.serialize(zone_to_test)
        self.assertEqual(created_now, serialized['updated_at'])

    def test_facets(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=False
        )

        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine

        mock_engine.search.return_value = {
            'aggregations': {
                'status': {
                    'buckets': [{'key': 'pending', 'doc_count': 2}]
                },
                'type': {
                    'buckets': [{'key': 'PRIMARY', 'doc_count': 2}]
                }
            },
            'hits': {'total': 2}
        }

        facets, _ = self.plugin.get_facets(fake_request.context)

        status_facet = list(filter(lambda f: f['name'] == 'status', facets))[0]
        expected_status = {
            'name': 'status',
            'options': [{'key': 'pending', 'doc_count': 2}],
            'type': 'string'
        }
        self.assertEqual(expected_status, status_facet)

        expected_agg_query = {
            'aggs': dict(unit_test_utils.simple_facet_field_agg(name)
                         for name in ('status', 'type')),
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {ROLE_USER_FIELD: 'user'}},
                            'should': [{'term': {'project_id': TENANT1}}],
                            'minimum_should_match': 1
                        }
                    }
                }
            }
        }
        mock_engine.search.assert_called_with(
            index=self.plugin.alias_name_search,
            doc_type=self.plugin.get_document_type(),
            body=expected_agg_query,
            ignore_unavailable=True,
            size=0
        )

    def test_serialize(self):
        serialized = self.plugin.serialize(self.zone1)
        self.assertEqual(TENANT1, serialized['project_id'])

    def test_delete_zone(self):
        delete_event = (
            "dns.zone.delete",
            {
                "shard": 776,
                "minimum": 3600,
                "ttl": 3600,
                "serial": 1459409054,
                "deleted_at": None,
                "id": "3081593e-10ca-408c-af77-1397e689c177",
                "parent_zone_id": None,
                "retry": 600,
                "transferred_at": None,
                "version": 10,
                "type": "PRIMARY",
                "email": "foo@example.org",
                "status": "PENDING",
                "description": None,
                "deleted": "0",
                "updated_at": "2016-03-31T17:39:05.000000",
                "expire": 86400,
                "masters": [],
                "name": "myzone.net.",
                "tenant_id": "80264096ac454d3d904002491fafe2ec",
                "created_at": "2016-03-31T05:30:02.000000",
                "pool_id": "794ccc2c-d751-44fe-b57f-8894c9f5c842",
                "refresh": 3588,
                "delayed_notify": False,
                "action": "DELETE",
                "attributes": []
            },
            "2016-03-31 17:39:05.495730"
        )

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        with mock.patch.object(self.plugin.index_helper,
                               'delete_document') as mock_zone_delete:
            with mock.patch.object(
                    self.plugin.child_plugins[0].index_helper,
                    'delete_documents') as mock_recordset_delete:
                with mock.patch(
                        'elasticsearch.helpers.scan') as mock_scan:
                    # mock scan method to return child recordset
                    mock_scan.return_value = [{'_id': 'recordset1'}]
                    type_handler = event_handlers.get(delete_event[0], None)
                    result = type_handler(*delete_event)
                    self.assertEqual(1, mock_zone_delete.call_count)
                    self.assertEqual(1, mock_recordset_delete.call_count)
                    self.assertEqual(2, len(result))
                    self.check_item(result[0],
                                    pipeline.DeleteItem,
                                    delete_event[0],
                                    delete_event[1],
                                    designate.recordsets.RecordSetIndex
                                    )
                    self.assertEqual("recordset1", result[0].doc_id)
                    self.check_item(result[1],
                                    pipeline.DeleteItem,
                                    delete_event[0],
                                    delete_event[1],
                                    designate.zones.ZoneIndex)
                    self.assertEqual(
                        "3081593e-10ca-408c-af77-1397e689c177",
                        result[1].doc_id)

    def test_update_zone(self):
        update_event = (
            "dns.zone.update",
            {
                "shard": 776,
                "minimum": 3600,
                "ttl": 4800,
                "serial": 1459402269,
                "deleted_at": None,
                "id": "3081593e-10ca-408c-af77-1397e689c177",
                "parent_zone_id": None,
                "retry": 600,
                "transferred_at": None,
                "version": 2,
                "type": "PRIMARY",
                "email": "foo@example.org",
                "status": "PENDING",
                "description": None,
                "deleted": "0",
                "updated_at": "2016-03-31T05:31:10.000000",
                "expire": 86400,
                "masters": [],
                "name": "myzone.net.",
                "tenant_id": "80264096ac454d3d904002491fafe2ec",
                "created_at": "2016-03-31T05:30:02.000000",
                "pool_id": "794ccc2c-d751-44fe-b57f-8894c9f5c842",
                "refresh": 3588,
                "delayed_notify": False,
                "action": "UPDATE",
                "attributes": []
            },
            "2016-03-31 05:31:10.232924"
        )

        handler = self.plugin.get_notification_handler()
        event_handlers = handler.get_event_handlers()

        with mock.patch.object(self.plugin.index_helper,
                               'save_document') as mock_save:
            result = event_handlers.get(update_event[0])(*update_event)
            self.assertEqual(1, mock_save.call_count)
            self.check_item(result,
                            pipeline.IndexItem,
                            update_event[0],
                            update_event[1],
                            designate.zones.ZoneIndex
                            )

    def test_initial_create_zone(self):
        create_event = (
            "dns.zone.create",
            {
                "shard": 776,
                "minimum": 3600,
                "ttl": 3600,
                "serial": 1459402202,
                "deleted_at": None,
                "id": "3081593e-10ca-408c-af77-1397e689c177",
                "parent_zone_id": None,
                "retry": 600,
                "transferred_at": None,
                "version": 1,
                "type": "PRIMARY",
                "email": "foo@example.org",
                "status": "PENDING",
                "description": None,
                "deleted": "0",
                "updated_at": None,
                "expire": 86400,
                "masters": [],
                "name": "myzone.net.",
                "tenant_id": "80264096ac454d3d904002491fafe2ec",
                "created_at": "2016-03-31T05:30:02.000000",
                "pool_id": "794ccc2c-d751-44fe-b57f-8894c9f5c842",
                "refresh": 3588,
                "delayed_notify": False,
                "action": "CREATE",
                "attributes": []
            },
            "2016-03-31 05:30:02.692222"
        )
        handler = self.plugin.get_notification_handler()
        with mock.patch(
                'searchlight.elasticsearch.plugins.designate._get_recordsets'
        ) as mock_list:
            with mock.patch.object(handler.index_helper,
                                   'save_documents') as mock_save:
                with mock.patch.object(handler.recordset_helper,
                                       'save_documents') as mock_rs_save:
                    mock_list.return_value = [
                        _recordset_fixture(
                            "1",
                            "3081593e-10ca-408c-af77-1397e689c177",
                            "80264096ac454d3d904002491fafe2ec",
                            name='www.test.com.',
                            type='A',
                            records=['192.0.2.1'])
                    ]
                    result = handler.process({},
                                             "central.ubunt",
                                             create_event[0],
                                             create_event[1],
                                             {'timestamp': create_event[2]}
                                             )
                    self.assertEqual(1, mock_save.call_count)
                    self.assertEqual(1, mock_rs_save.call_count)
                    self.assertEqual(len(result), 2)
                    self.check_item(result[0],
                                    pipeline.IndexItem,
                                    create_event[0],
                                    create_event[1],
                                    designate.zones.ZoneIndex)
                    self.assertEqual("3081593e-10ca-408c-af77-1397e689c177",
                                     result[0].doc_id)
                    self.check_item(result[1],
                                    pipeline.IndexItem,
                                    create_event[0],
                                    create_event[1],
                                    designate.recordsets.RecordSetIndex)
                    self.assertEqual("1", result[1].doc_id)



# Copyright (c) 2015-2016 Hewlett-Packard Enterprise Development Company, L.P.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

GLANCE_IMAGE = "OS::Glance::Image"
GLANCE_METADEF = "OS::Glance::Metadef"

CINDER_SNAPSHOT = "OS::Cinder::Snapshot"
CINDER_VOLUME = "OS::Cinder::Volume"

DESIGNATE_ZONE = "OS::Designate::Zone"
DESIGNATE_RECORDSET = "OS::Designate::RecordSet"

NEUTRON_NETWORK = "OS::Neutron::Net"
NEUTRON_PORT = "OS::Neutron::Port"
NEUTRON_SUBNET = "OS::Neutron::Subnet"
NEUTRON_ROUTER = "OS::Neutron::Router"
NEUTRON_SUBNET_POOL = "OS::Neutron::SubnetPool"
NEUTRON_FLOATINGIP = "OS::Neutron::FloatingIP"
NEUTRON_SECURITY_GROUP = "OS::Neutron::SecurityGroup"

NOVA_SERVER = "OS::Nova::Server"
NOVA_SERVERGROUP = "OS::Nova::ServerGroup"
NOVA_HYPERVISOR = "OS::Nova::Hypervisor"
NOVA_FLAVOR = "OS::Nova::Flavor"
NOVA_AVAILABILITY_ZONE = "OS::Nova::AvailabilityZone"
NOVA_SECURITY_GROUP = "OS::Nova::SecurityGroup"

SWIFT_ACCOUNT = "OS::Swift::Account"
SWIFT_CONTAINER = 'OS::Swift::Container'
SWIFT_OBJECT = 'OS::Swift::Object'

KEYSTONE_PROJECT = "OS::Keystone::Project"
KEYSTONE_USER = "OS::Keystone::User"

IRONIC_NODE = "OS::Ironic::Node"
IRONIC_PORT = "OS::Ironic::Port"
IRONIC_CHASSIS = "OS::Ironic::Chassis"



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import webob

from searchlight.api.v1 import search
from searchlight.common import utils
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils

from searchlight.api import policy


REPO_SEARCH = 'searchlight.elasticsearch.CatalogSearchRepo.search'


class TestSearchPolicy(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSearchPolicy, self).setUp()
        self.enforcer = policy.Enforcer()
        self.enforcer.load_rules()

    def test_admin_or_owner(self):
        """Since this a commonly used rule, check that it works"""
        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'admin_or_owner': 'role:admin or project_id:%(project_id)s',
            'test_rule': 'rule:admin_or_owner'
        }))

        request = unit_test_utils.get_fake_request(is_admin=False)
        self.assertTrue(self.enforcer.check(request.context, 'test_rule',
                                            request.context.policy_target))

        request = unit_test_utils.get_fake_request(is_admin=False)
        self.assertFalse(self.enforcer.check(request.context, 'test_rule',
                                             {}))

        request = unit_test_utils.get_fake_request(is_admin=True)
        self.assertTrue(self.enforcer.check(request.context, 'test_rule',
                                            {}))

    def test_context_policy_target(self):
        request = unit_test_utils.get_fake_request()
        expected = {'user_id': unit_test_utils.SOMEUSER,
                    'project_id': unit_test_utils.SOMETENANT,
                    'tenant_id': unit_test_utils.SOMETENANT}
        self.assertEqual(expected, request.context.policy_target)

    def test_search_policy(self):
        request = unit_test_utils.get_fake_request()
        search_repo = mock.Mock()
        with mock.patch.object(self.enforcer, 'enforce') as mock_enforce:
            repo = policy.CatalogSearchRepoProxy(search_repo, request.context,
                                                 self.enforcer)
            repo.search()
            mock_enforce.assert_called_with(request.context,
                                            'search:query',
                                            request.context.policy_target)

    def test_plugin_policy(self):
        request = unit_test_utils.get_fake_request()
        search_repo = mock.Mock()
        with mock.patch.object(self.enforcer, 'enforce') as mock_enforce:
            repo = policy.CatalogSearchRepoProxy(search_repo, request.context,
                                                 self.enforcer)
            repo.plugins_info()
            mock_enforce.assert_called_with(request.context,
                                            'search:plugins_info',
                                            request.context.policy_target)

    def test_facet_policy(self):
        request = unit_test_utils.get_fake_request()
        search_repo = mock.Mock()
        with mock.patch.object(self.enforcer, 'enforce') as mock_enforce:
            repo = policy.CatalogSearchRepoProxy(search_repo, request.context,
                                                 self.enforcer)
            repo.facets()
            mock_enforce.assert_called_with(request.context,
                                            'search:facets',
                                            request.context.policy_target)

    @mock.patch('searchlight.api.v1.search.' +
                'RequestDeserializer._get_request_body')
    def test_search_resource_policy_checks(self, mock_request_body):
        request = unit_test_utils.get_fake_request()
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        with mock.patch.object(self.enforcer, 'check') as mock_enforce:
            mock_request_body.return_value = {'type': 'OS::Nova::Server'}
            search_deserializer.search(request)
            mock_enforce.assert_called_with(request.context,
                                            'resource:OS::Nova::Server',
                                            request.context.policy_target)

    def test_plugins_info_resource_policy(self):
        request = unit_test_utils.get_fake_request()
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        with mock.patch.object(self.enforcer, 'check') as mock_enforce:
            search_deserializer.plugins_info(request)
            self.assertIn(mock.call(request.context,
                                    'resource:OS::Nova::Server',
                                    request.context.policy_target),
                          mock_enforce.mock_calls)
            self.assertIn(mock.call(request.context,
                                    'resource:OS::Glance::Image',
                                    request.context.policy_target),
                          mock_enforce.mock_calls)
            self.assertEqual(len(utils.get_search_plugins()),
                             len(mock_enforce.call_args_list))

    def test_facets_resource_policy(self):
        request = unit_test_utils.get_fake_request()
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        with mock.patch.object(self.enforcer, 'check') as mock_enforce:
            search_deserializer.facets(request)
            self.assertIn(mock.call(request.context,
                                    'resource:OS::Nova::Server',
                                    request.context.policy_target),
                          mock_enforce.mock_calls)
            self.assertIn(mock.call(request.context,
                                    'resource:OS::Glance::Image',
                                    request.context.policy_target),
                          mock_enforce.mock_calls)
            self.assertEqual(len(utils.get_search_plugins()),
                             len(mock_enforce.call_args_list))

    def test_resource_policy_allowed(self):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)
        types = ['OS::Glance::Image', 'OS::Nova::Server']

        self.assertEqual(
            types,
            search_deserializer._filter_types_by_policy(request.context, types)
        )

        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Glance::Image': '',
            'resource:OS::Nova::Server': ''
        }))

        self.assertEqual(
            types,
            search_deserializer._filter_types_by_policy(request.context, types)
        )

    def test_resource_policy_disallow_non_admin(self):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)
        types = ['OS::Glance::Image', 'OS::Nova::Server',
                 'OS::Glance::Metadef']

        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Glance::Image': 'role:admin'
        }))
        filtered_types = search_deserializer._filter_types_by_policy(
            request.context, types)
        self.assertEqual(set(['OS::Nova::Server', 'OS::Glance::Metadef']),
                         set(filtered_types))

        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Nova::Server': 'role:admin'
        }))

        filtered_types = search_deserializer._filter_types_by_policy(
            request.context, types)
        self.assertEqual(['OS::Glance::Metadef'], filtered_types)

    def test_resource_policy_allows_admin(self):
        request = unit_test_utils.get_fake_request(is_admin=True)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)
        types = ['OS::Glance::Image', 'OS::Nova::Server']

        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Glance::Image': 'role:admin',
            'resource:OS::Nova::Server': 'role:admin'
        }))

        filtered_types = search_deserializer._filter_types_by_policy(
            request.context, types)
        self.assertEqual(types, filtered_types)

    def test_resource_policy_disallow(self):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)
        types = ['OS::Glance::Image', 'OS::Nova::Server']

        # And try disabling access for everyone
        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Glance::Image': '!'
        }))

        self.assertEqual(
            ['OS::Nova::Server'],
            search_deserializer._filter_types_by_policy(request.context, types)
        )

        # Same for admin
        request = unit_test_utils.get_fake_request(is_admin=False)
        self.assertEqual(
            ['OS::Nova::Server'],
            search_deserializer._filter_types_by_policy(request.context, types)
        )

    def test_policy_all_disallowed(self):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        types = ['OS::Glance::Image']
        self.enforcer.add_rules(policy.policy.Rules.from_dict({
            'resource:OS::Glance::Image': '!',
            'resource:OS::Nova::Server': '!',
        }))

        expected_message = (
            "There are no resource types accessible to you to serve "
            "your request. You do not have access to the following "
            "resource types: OS::Glance::Image")
        self.assertRaisesRegex(
            webob.exc.HTTPForbidden, expected_message,
            search_deserializer._filter_types_by_policy,
            request.context, types)

        types = ['OS::Glance::Image', 'OS::Nova::Server']
        expected_message = (
            "There are no resource types accessible to you to serve "
            "your request. You do not have access to the following "
            "resource types: OS::Glance::Image, OS::Nova::Server")
        self.assertRaisesRegex(
            webob.exc.HTTPForbidden, expected_message,
            search_deserializer._filter_types_by_policy,
            request.context, types)

    def test_search_service_policies(self):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        types = ['OS::Glance::Image', 'OS::Nova::Server']
        glance_enforce = mock.Mock()
        nova_enforce = mock.Mock()
        glance_enforce.enforce.return_value = False
        nova_enforce.enforce.return_value = True
        service_enforcers = {
            'image': glance_enforce,
            'compute': nova_enforce
        }

        expect_creds = {
            'tenant_id': request.context.tenant,
            'project_id': request.context.tenant,
            'user_id': request.context.user,
            'roles': ['member'],
            'is_admin_project': True,
            'is_admin': False,
            'domain_id': None,
            'user_domain_id': None,
            'project_domain_id': None,
            'service_user_id': None,
            'service_user_domain_id': None,
            'service_project_id': None,
            'service_project_domain_id': None,
            'service_roles': [],
            'system_scope': None
        }

        fake_target = {
            'user_id': request.context.user,
            'project_id': request.context.tenant,
            'tenant_id': request.context.tenant
        }

        with mock.patch('searchlight.service_policies._get_enforcers',
                        return_value=service_enforcers):
            filtered_types = search_deserializer._filter_types_by_policy(
                request.context, types)
            self.assertEqual(['OS::Nova::Server'], filtered_types)
            glance_enforce.enforce.assert_called_with(
                'get_images', fake_target, expect_creds)
            nova_enforce.enforce.assert_called_with(
                'os_compute_api:servers:index', fake_target, expect_creds)

    @mock.patch('searchlight.api.v1.search.' +
                'RequestDeserializer._get_request_body')
    def test_aggregation_policy(self, mock_request_body):
        request = unit_test_utils.get_fake_request(is_admin=False)
        search_deserializer = search.RequestDeserializer(
            utils.get_search_plugins(),
            policy_enforcer=self.enforcer)

        with mock.patch.object(self.enforcer, 'enforce') as mock_enforce:
            mock_request_body.return_value = {
                'query': {'match_all': {}},
                'aggregations': {'terms': {'field': 'some_field'}}
            }
            search_deserializer.search(request)
            mock_enforce.assert_called_with(request.context,
                                            'search:query:aggregations',
                                            request.context.policy_target)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
import operator

import glanceclient.exc
import six

from searchlight.elasticsearch.plugins import openstack_clients
from searchlight.elasticsearch.plugins import utils

LOG = logging.getLogger(__name__)


def _get_image_members(image):
    if image['visibility'] in ['public', 'private']:
        return []

    try:
        g_client = openstack_clients.get_glanceclient()
        members = g_client.image_members.list(image['id'])
        return members
    except glanceclient.exc.Unauthorized:
        LOG.warning("Could not list image members for %s; forbidden" %
                    image['id'])
        return []


def _normalize_visibility(image_doc):
    """Adjust for a difference in v1 (and notifications) versus v2 where
    v1 uses 'is_public' (bool) and v2 uses 'visibility'
    (public/private/shared).
    Normalize everything to the v2 model.
    """
    if image_doc.get('visibility'):
        return
    is_public = image_doc.pop('is_public', None)
    if is_public is not None:
        image_doc['visibility'] = 'public' if is_public else 'private'


def serialize_glance_image(image):

    # If we're being asked to index an ID, retrieve the full image information
    if isinstance(image, six.text_type):
        g_client = openstack_clients.get_glanceclient()
        image = g_client.images.get(image)

    members = _get_image_members(image)

    fields_to_ignore = ['schema', 'file', 'locations']

    document = {k: v for k, v in image.items() if k not in fields_to_ignore}

    # Only snapshot images return image_type field.
    # Default to 'image' for non-snapshot image.
    if 'image_type' not in document:
        document['image_type'] = 'image'

    if 'project_id' not in document:
        document['project_id'] = document['owner']

    members = list(members)
    older_glance_version = False
    # Make it backward compatible with older glance version
    if members and len(members) > 0 and "deleted" in members[0]:
        older_glance_version = True

    if older_glance_version:
        document['members'] = [
            member['member'] for member in members
            if (member['status'] == 'accepted' and member['deleted'] == 0)]
    else:
        document['members'] = [
            member['member_id'] for member in members
            if (member['status'] == 'accepted')]

    return document


def serialize_glance_notification(note):
    """Notifications are different in several ways from v2 API images"""
    _normalize_visibility(note)

    # If an image is deleted we wouldn't have it in the index, so no need
    # to retain these fields
    for attr in ('deleted', 'deleted_at'):
        note.pop(attr, None)

    for k, v in note.pop('properties', {}).items():
        note[k] = v
    # 'Members' will be set in serialize_glance_image
    return serialize_glance_image(note)


def serialize_glance_image_members(image, payload):
    member_id = payload['member_id']
    if member_id in image['members']:
        if payload['status'].lower() != "accepted" or payload['deleted']:
            image['members'].remove(member_id)
    else:
        if payload['status'].lower() == "accepted" and \
                (not payload['deleted']):
            image['members'].append(member_id)
    return image


def serialize_glance_metadef_ns(metadef_namespace):
    def _serialize_tag(tag):
        return tag["name"]

    def _serialize_property(name, property):
        serialized_prop = copy.deepcopy(property)
        serialized_prop['name'] = name
        if 'default' in serialized_prop:
            serialized_prop['default'] = str(serialized_prop['default'])
        if 'enum' in serialized_prop:
            serialized_prop['enum'] = sorted(map(str, serialized_prop['enum']))

        return serialized_prop

    def _serialize_object(obj):
        serialized_obj = {
            'name': obj['name'],
            'description': obj['description']
        }
        serialized_obj['properties'] = sorted([
            _serialize_property(name, property)
            for name, property in obj.get('properties', {}).items()
        ], key=operator.itemgetter('name'))
        return serialized_obj

    def _serialize_res_type(rt):
        return {
            'name': rt['name']
        }

    # TODO(sjmc7): test this better
    incomplete = 'objects' not in metadef_namespace or \
                 'properties' not in metadef_namespace or \
                 'tags' not in metadef_namespace
    if incomplete:
        LOG.debug("Retrieving metadef namespace '%s'",
                  metadef_namespace['namespace'])
        g_client = openstack_clients.get_glanceclient()
        metadef_namespace = g_client.metadefs_namespace.get(
            metadef_namespace['namespace'])

    # The CIS code specifically serialized some fields rather than indexing
    # everything; do the same.
    namespace_fields = ('namespace', 'display_name', 'description',
                        'visibility', 'owner', 'protected', 'created_at',
                        'updated_at')
    document = {f: metadef_namespace.get(f, None) for f in namespace_fields}
    document['id'] = document['namespace']
    if document['display_name']:
        document['name'] = document['display_name']
    else:
        document['name'] = document['namespace']

    # Add project_id as a copy of 'owner'
    if 'project_id' not in document:
        document['project_id'] = document['owner']

    document['tags'] = sorted([
        _serialize_tag(tag) for tag in metadef_namespace.get('tags', [])])
    document['properties'] = sorted([
        _serialize_property(name, property)
        for name, property in metadef_namespace.get('properties', {}).items()
    ], key=operator.itemgetter('name'))
    document['objects'] = sorted([
        _serialize_object(obj) for obj in metadef_namespace.get('objects', [])
    ], key=operator.itemgetter('name'))
    document['resource_types'] = sorted([
        _serialize_res_type(rt)
        for rt in metadef_namespace.get('resource_type_associations', [])
    ], key=operator.itemgetter('name'))

    utils.normalize_date_fields(document)

    return document



# Copyright (c) 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re

"""
Guidelines for writing new hacking checks

 - Use only for Searchlight-specific tests. OpenStack general tests
   should be submitted to the common 'hacking' module.
 - Pick numbers in the range SL3xx. Find the current test with
   the highest allocated number and then pick the next value.
   If nova has an N3xx code for that test, use the same number.
 - Keep the test method code in the source file ordered based
   on the SL3xx value.
 - List the new rule in the top level HACKING.rst file
 - Add test cases for each new rule to searchlight/tests/test_hacking.py

"""


asse_trueinst_re = re.compile(
    r"(.)*assertTrue\(isinstance\((\w|\.|\'|\"|\[|\])+, "
    "(\w|\.|\'|\"|\[|\])+\)\)")
asse_equal_type_re = re.compile(
    r"(.)*assertEqual\(type\((\w|\.|\'|\"|\[|\])+\), "
    "(\w|\.|\'|\"|\[|\])+\)")
asse_equal_end_with_none_re = re.compile(
    r"(.)*assertEqual\((\w|\.|\'|\"|\[|\])+, None\)")
asse_equal_start_with_none_re = re.compile(
    r"(.)*assertEqual\(None, (\w|\.|\'|\"|\[|\])+\)")
unicode_func_re = re.compile(r"(\s|\W|^)unicode\(")
doubled_words_re = re.compile(
    r"\b(then?|[iao]n|i[fst]|but|f?or|at|and|[dt]o)\s+\1\b")
translated_logs = re.compile(
    r"(.)*LOG.(critical|debug|error|exception|info|warning)\(\s*_\(")


def assert_true_instance(logical_line):
    """Check for assertTrue(isinstance(a, b)) sentences

    SL316
    """
    if asse_trueinst_re.match(logical_line):
        yield (0, "SL316: assertTrue(isinstance(a, b)) sentences not allowed")


def assert_equal_type(logical_line):
    """Check for assertEqual(type(A), B) sentences

    SL317
    """
    if asse_equal_type_re.match(logical_line):
        yield (0, "SL317: assertEqual(type(A), B) sentences not allowed")


def assert_equal_none(logical_line):
    """Check for assertEqual(A, None) or assertEqual(None, A) sentences

    SL318
    """
    res = (asse_equal_start_with_none_re.match(logical_line) or
           asse_equal_end_with_none_re.match(logical_line))
    if res:
        yield (0, "SL318: assertEqual(A, None) or assertEqual(None, A) "
               "sentences not allowed")


def no_translate_logs(logical_line, filename):
    dirs = [
        "searchlight/api",
        "searchlight/cmd",
        "searchlight/common",
        "searchlight/elasticsearch",
        "searchlight/tests",
    ]

    if max([name in filename for name in dirs]):
        if translated_logs.match(logical_line):
            yield(0, "SL319: Don't translate logs")


def no_direct_use_of_unicode_function(logical_line):
    """Check for use of unicode() builtin

    SL320
    """
    if unicode_func_re.match(logical_line):
        yield(0, "SL320: Use six.text_type() instead of unicode()")


def check_no_contextlib_nested(logical_line):
    msg = ("SL327: contextlib.nested is deprecated since Python 2.7. See "
           "https://docs.python.org/2/library/contextlib.html#contextlib."
           "nested for more information.")
    if ("with contextlib.nested(" in logical_line or
            "with nested(" in logical_line):
        yield(0, msg)


def check_doubled_words(physical_line, filename):
    """Check for the common doubled-word typos

    N343
    """
    msg = ("N343: Doubled word '%(word)s' typo found")

    match = re.search(doubled_words_re, physical_line)

    if match:
        return (0, msg % {'word': match.group(1)})


def factory(register):
    register(assert_true_instance)
    register(assert_equal_type)
    register(assert_equal_none)
    register(no_translate_logs)
    register(no_direct_use_of_unicode_function)
    register(check_no_contextlib_nested)
    register(check_doubled_words)



# Copyright 2015 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import calendar
import logging
from oslo_config import cfg
from oslo_config import types
from oslo_utils import encodeutils
from oslo_utils import timeutils
import re
import six

from searchlight.common import exception
import searchlight.elasticsearch
from searchlight.elasticsearch.plugins import helper
from searchlight.elasticsearch.plugins import utils
from searchlight.elasticsearch import ROLE_USER_FIELD
from searchlight.i18n import _
from searchlight import plugin


LOG = logging.getLogger(__name__)

resource_group_reg = re.compile(r'^[a-z0-9][a-z0-9_]*$')
indexer_opts = [
    cfg.Opt('resource_group_name', default="searchlight",
            help="The default base name for accessing Elasticsearch",
            type=types.String(regex=resource_group_reg)),
    cfg.StrOpt('notifications_topic', default="notifications",
               help="The default messaging notifications topic"),
    cfg.BoolOpt('mapping_use_doc_values', default=True,
                help='Use doc_values for mapped fields where applicable.'
                     'Allows lower memory usage at the cost of some disk '
                     'space. Recommended, especially in large deployments.'),
    cfg.BoolOpt('include_region_name', default=False,
                help='Whether or not to include region_name as a mapping '
                     'field and in documents. The value will be determined '
                     'from service_credentials and override_region_name for '
                     'each plugin.'),
    cfg.ListOpt('publishers',
                help='The global publishers configuration, '
                     'plugin can override this by setting their own publishers'
                )
]

CONF = cfg.CONF
CONF.register_opts(indexer_opts, group='resource_plugin')


@six.add_metaclass(abc.ABCMeta)
class IndexBase(plugin.Plugin):
    NotificationHandlerCls = None

    def __init__(self):
        self.options = cfg.CONF[self.get_config_group_name()]

        self.engine = searchlight.elasticsearch.get_api()
        self.document_type = self.get_document_type()

        # This will be populated at load time.
        self.child_plugins = []
        self.parent_plugin = None

    @property
    def index_helper(self):
        if not getattr(self, '_index_helper', None):
            self._index_helper = helper.IndexingHelper(self)
        return self._index_helper

    @property
    def name(self):
        return "%s-%s" % (self.resource_group_name, self.document_type)

    @property
    def mapping_use_doc_values(self):
        """Set to true to disable doc_values for a plugin's entire mapping"""
        if self.options.mapping_use_doc_values is not None:
            return self.options.mapping_use_doc_values
        else:
            return cfg.CONF.resource_plugin.mapping_use_doc_values

    @property
    def resource_group_name(self):
        if not getattr(self, '_group_name', None):
            if self.parent_plugin:
                self._group_name = self.parent_plugin.resource_group_name
                if self.options.resource_group_name is not None and  \
                        self.options.resource_group_name != self._group_name:
                    LOG.warning(
                        "Overriding resource_group for %(plugin)s because it "
                        "differs from parent plugin %(parent)s resource_group "
                        "%(resource_group)s" %
                        {"plugin": self.document_type,
                         "parent": self.parent_plugin.document_type,
                         "resource_group": self._group_name}
                    )
            else:
                if self.options.resource_group_name is not None:
                    self._group_name = self.options.resource_group_name
                elif cfg.CONF.resource_plugin.resource_group_name is not None:
                    self._group_name = \
                        cfg.CONF.resource_plugin.resource_group_name
                else:
                    self._group_name = "searchlight"
        return self._group_name

    @property
    def alias_name_listener(self):
        return "%s-listener" % self.resource_group_name

    @property
    def alias_name_search(self):
        return "%s-search" % self.resource_group_name

    @abc.abstractproperty
    def resource_allowed_policy_target(self):
        """Should return the policy target that this plugin's related service
        uses to determine who has API access to list resources. Can return None
        if this service doesn't support policy.
        """
        return None

    @abc.abstractproperty
    def service_type(self):
        """Should match the keystone catalog service type for a resource"""
        return None

    def prepare_index(self, index_name):
        """Prepare a new index for usage with this listener. We need to be
           called immediately after a new index is created, but before it
           gets associated with an alias. Prepping means we will add
           the settings and mapping for this listener's document type.
        """
        if self.parent_plugin_type():
            LOG.debug(
                "Skipping index prep for %(doc_type)s; will be handled by "
                "parent (%(parent_type)s)" %
                {"doc_type": self.document_type,
                 "parent_type": self.parent_plugin_type()})
            return

        self.check_mapping_sort_fields()
        for child_plugin in self.child_plugins:
            child_plugin.check_mapping_sort_fields()

        # Prepare the new index for this document type.
        self.setup_index_mapping(index_name=index_name)

    def setup_index_settings(self, index_name):
        """Update index settings. """
        index_settings = self.get_settings()
        if index_settings:
            self.engine.indices.put_settings(body=index_settings,
                                             index=index_name)

    def setup_index_mapping(self, index_name):
        """Update index document mapping."""
        # Using 'reversed' because in e-s 2.x, child mappings must precede
        # their parents, and the parent will be the first element
        for doc_type, mapping in self.get_full_mapping():
            self.engine.indices.put_mapping(index=index_name,
                                            doc_type=doc_type,
                                            body=mapping)

    def index_initial_data(self, index_name=None):
        """Insert all objects from a plugin (generally by API requests to its
        service) into an index assumed to have been created with
        prepare_index. If index_name is not set, the searchlight-listener alias
        will be used instead.
        """
        object_list = self.get_objects()
        documents = []
        versions = []
        for obj in object_list:
            document = self.serialize(obj)
            documents.append(document)
            if self.NotificationHandlerCls:
                version = self.NotificationHandlerCls.get_version(document)
            else:
                version = NotificationBase.get_version(document)
            versions.append(version)
        self.index_helper.save_documents(documents, versions=versions,
                                         index=index_name)

    def get_facets(self, request_context, all_projects=False, limit_terms=0,
                   include_fields=True, exclude_options=False):
        """Get facets available for searching, in the form of a list of
        dicts with keys "name", "type" and optionally "options" if a field
        should have discreet allowed values. If include_fields is false,
        only the total doc count will be requested.
        """
        exclude_facets = self.facets_excluded
        is_admin = request_context.is_admin

        def include_facet(name):
            if name not in exclude_facets:
                return True

            if is_admin and exclude_facets[name]:
                return True

            return False

        def get_facets_for(property_mapping, meta_mapping, prefix='',
                           inside_nested=None):
            mapping_facets = []
            for name, properties in property_mapping.items():
                property_type = properties.get('type', 'object')
                if property_type in ('nested', 'object'):
                    if include_facet(prefix + name):
                        is_nested = property_type == 'nested'
                        mapping_facets.extend(
                            get_facets_for(properties['properties'],
                                           meta_mapping,
                                           "%s%s." % (prefix, name),
                                           inside_nested=is_nested))
                else:
                    indexed = properties.get('index', None) != 'no'
                    if indexed and include_facet(name):
                        facet_name = prefix + name
                        mapping_facet = {
                            'name': facet_name,
                            'type': property_type
                        }

                        # If we're inside either an object or nested object,
                        # add that to the mapping too
                        if inside_nested is not None:
                            mapping_facet['nested'] = inside_nested

                        # Add raw field if it's an analyzed field,
                        # aggregation on analyzed string field doesn't
                        # work well
                        if (properties['type'] == 'string' and
                                properties.get('index') != 'not_analyzed' and
                                'raw' in properties.get('fields', {})):
                            mapping_facet['facet_field'] = facet_name + '.raw'

                        # Plugin can specify _meta mapping to link an id with a
                        # resource type, which can be used by Client program/UI
                        # to GET more information from the resource type.
                        # See https://www.elastic.co/guide/en/elasticsearch/
                        # reference/2.1/mapping-meta-field.html
                        if facet_name in meta_mapping:
                            mapping_facet.update(meta_mapping[facet_name])

                        if (self.get_parent_id_field() and
                                name == self.get_parent_id_field()):
                            mapping_facet['parent'] = True
                            if 'resource_type' not in mapping_facet:
                                mapping_facet['resource_type'] = \
                                    self.parent_plugin_type()

                        mapping_facets.append(mapping_facet)

            return mapping_facets

        facets = []
        if include_fields:
            facets = get_facets_for(self.get_mapping()['properties'],
                                    self.get_mapping().get('_meta', {}))

        # Don't retrieve facet terms for any excluded fields
        included_fields = set(f['name'] for f in facets)
        options_fields = set(self.facets_with_options) & included_fields

        raw_fields = dict([(f['name'], f['facet_field'])
                           for f in facets
                           if f.get('facet_field')])
        # If field has a raw field, use a tuple of field name and
        # raw field name
        facet_terms_for = [(field, raw_fields[field])
                           if field in raw_fields
                           else field
                           for field in options_fields]

        facet_terms, doc_count = self._get_facet_terms(
            facet_terms_for, request_context, all_projects,
            limit_terms=limit_terms, exclude_options=exclude_options)

        if include_fields:
            for facet in facets:
                if facet['name'] in facet_terms:
                    facet['options'] = facet_terms[facet['name']]

        return facets, doc_count

    @property
    def facets_excluded(self):
        """A map of {name: allow_admin} that indicate which
        fields should not be offered as facet options.
        """
        return {}

    @property
    def facets_with_options(self):
        """An iterable of facet names that support facet options"""
        return ()

    def _get_facet_terms(self, fields, request_context,
                         all_projects, limit_terms,
                         exclude_options=False):
        # Fields can be empty if there are no facet terms desired,
        # but we will run a size=0 search to get the doc count. If the
        # user does not want the options associated with the facet fields,
        # do not aggregate. This is controlled with a parameter to the
        # API call.
        body = {}
        term_aggregations = {}

        if fields:
            mapping = self.get_mapping()['properties']
            # Nested fields will be all the mapped fields with type=='nested'
            # Not tremendously keen on the way this is structured but the
            # nested fields are a bit special case
            nested_fields = [name
                             for name, properties in mapping.items()
                             if properties['type'] == 'nested']
            if not exclude_options:
                term_aggregations = utils.get_facets_query(fields,
                                                           nested_fields,
                                                           limit_terms)
            if term_aggregations:
                body['aggs'] = term_aggregations

        role_filter = request_context.user_role_filter

        filter_query = {
            "bool": {
                "filter": {
                    "bool": {
                        "must": {
                            "term": {ROLE_USER_FIELD: role_filter}
                        }
                    }
                }
            }
        }

        # Add in the RBAC filters unless all_projects is requested
        if not (request_context.is_admin and all_projects):
            rbac_filters = self._get_rbac_field_filters(request_context)

            # minimum_should_match:1 is assumed in filter context,
            # but I'm including it explicitly so nobody spends an hour
            # scouring the documentation to check that is the case
            if rbac_filters:
                filter_query["bool"]["filter"]["bool"].update(
                    {"should": self._get_rbac_field_filters(request_context),
                     "minimum_should_match": 1})

        body['query'] = filter_query

        results = self.engine.search(
            index=self.alias_name_search,
            doc_type=self.get_document_type(),
            body=body,
            ignore_unavailable=True,
            size=0)

        agg_results = results.get('aggregations', {})
        doc_count = results['hits']['total']

        facet_terms = utils.transform_facets_results(
            agg_results,
            self.get_document_type())

        if term_aggregations and not agg_results:
            LOG.warning(
                "No aggregations found for %(resource_type)s. There may "
                "be a mapping problem." %
                {'resource_type': self.get_document_type()})
        return facet_terms, doc_count

    def check_mapping_sort_fields(self):
        """Check that fields that are expected to define a 'raw' field do so"""
        fields_needing_raw = searchlight.elasticsearch.RAW_SORT_FIELDS
        mapped_properties = self.get_mapping().get('properties', {})
        for field_name, field_mapping in mapped_properties.items():
            if field_name in fields_needing_raw:
                raw = field_mapping.get('fields', {}).get('raw', None)
                if not raw:
                    msg_vals = {"field_name": field_name,
                                "index_name": self.alias_name_listener,
                                "document_type": self.get_document_type()}
                    message = ("Field '%(field_name)s' for %(index_name)s/"
                               "%(document_type)s must contain a subfield "
                               "whose name is 'raw' for sorting." % msg_vals)
                    raise Exception(message)

    @abc.abstractmethod
    def get_objects(self):
        """Get list of all objects which will be indexed into search engine."""

    @abc.abstractmethod
    def serialize(self, obj):
        """Serialize database object into valid search engine document."""

    def get_document_id_field(self):
        """Whatever document field should be treated as the id. This field
        should also be mapped to _id in the elasticsearch mapping, though
        under role-based filtering, it will be modified to avoid clashes.
        """
        return "id"

    @classmethod
    def parent_plugin_type(cls):
        """If set, should be the resource type (canonical name) of
        the parent. Setting this for a plugin means that plugin cannot be
        initially indexed on its own, only as part of the parent indexing.
        """
        return None

    def get_parent_id_field(self):
        """Whatever field should be treated as the parent id. This is required
        for plugins with _parent definitions in their mappings. Documents to be
        indexed should contain this field.
        """
        return None

    @property
    def routing_field(self):
        """Whatever field should be treated as the routing value. Whatever is
        used must be available to all notifications on the child since
        index/update/delete operations will all require it. Additionally, if
        it is NOT the parent id, the parent plugin must be routed using the
        same field. This field is accessed AFTER serialization.
        """
        return None

    @property
    def include_region_name(self):
        return cfg.CONF.resource_plugin.include_region_name

    @property
    def region_name(self):
        """Returns a region name if enabled by config and configured in
        service_credentials. Adding it as per-plugin rather than global
        in case at some point per-plugin credentials are supported.
        """
        # Check the override in each plugin's options first, then use
        # whatever's in service_credentials
        region_name = self.options.override_region_name
        if region_name:
            return region_name
        region_name = getattr(CONF.service_credentials,
                              'os_region_name', None)
        if region_name:
            return [region_name]

    @property
    def enabled(self):
        return self.options.enabled

    @property
    def publishers(self):
        if self.options.publishers is not None:
            return self.options.publishers
        elif cfg.CONF.resource_plugin.publishers is not None:
            return cfg.CONF.resource_plugin.publishers

    @property
    def allow_admin_ignore_rbac(self):
        """If set for a plugin, an administrative query for all_projects will
        NOT skip RBAC filters.
        """
        return True

    @classmethod
    def get_document_type(cls):
        """Get name of the document type.

        This is in the format of OS::Service::Resource typically.
        """
        raise NotImplementedError()

    def register_parent(self, parent):
        if not self.parent_plugin:
            parent.child_plugins.append(self)
            self.parent_plugin = parent

    def get_query_filters(self, request_context, ignore_rbac=False):
        """Gets an index/type filter as es json filter dsl, and include
        plugin-specific RBAC filter terms unless `ignore_rbac` is True
        (for admin queries).
        """
        query_filter = {
            'indices': {
                'index': self.alias_name_search,
                'no_match_query': 'none',
                'query': {
                    'bool': {
                        'filter': {
                            'bool': {
                                'must': {
                                    'type': {
                                        'value': self.get_document_type()
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        if not (ignore_rbac and self.allow_admin_ignore_rbac):
            rbac_filters = self._get_rbac_field_filters(request_context)
            if rbac_filters:
                bool_clause = query_filter['indices']['query']['bool']
                bool_clause['filter']['bool'].update({
                    'should': rbac_filters,
                    'minimum_should_match': 1})

        return query_filter

    @abc.abstractmethod
    def _get_rbac_field_filters(self, request_context):
        """Return any RBAC field filters in a list to be injected into an
        indices query. Document type will be added. The filters will be used
        in the 'should' filter clause (i.e. OR).
        """

    def get_notification_handler(self):
        """Get the notification handler which implements NotificationBase."""
        if self.NotificationHandlerCls:
            return self.NotificationHandlerCls(self.index_helper,
                                               self.options)
        return None

    def filter_result(self, hit, request_context):
        """Filter each outgoing search result; document in hit['_source'].
        """
        if 'highlight' in hit:
            # We need to clean up any fields 'leaked' by highlighting.
            # As a reminder since this is a rare case, the 'highlight'
            # dict is a peer of '_source' in the overall hit structure.
            hit['highlight'].pop(ROLE_USER_FIELD, None)

    @abc.abstractmethod
    def get_mapping(self):
        """Get an index mapping."""

    def get_full_mapping(self):
        """Gets the full mapping doc for this type, including children. This
        returns a child-first (depth-first) generator.
        """
        # Assemble the children!
        for child_plugin in self.child_plugins:
            for plugin_type, mapping in child_plugin.get_full_mapping():
                yield plugin_type, mapping

        type_mapping = self.get_mapping()

        # Add common mapping fields
        # ROLE_USER_FIELD is required for RBAC by all plugins
        type_mapping['properties'][ROLE_USER_FIELD] = {
            'type': 'string',
            'index': 'not_analyzed',
            'include_in_all': False
        }

        # Add region name
        if self.include_region_name:
            type_mapping['properties']['region_name'] = {
                'type': 'string',
                'index': 'not_analyzed',
            }

        if 'updated_at' not in type_mapping['properties'].keys():
            type_mapping['properties']['updated_at'] = {'type': 'date'}

        if self.mapping_use_doc_values:
            helper.IndexingHelper.apply_doc_values(type_mapping)

        expected_parent_type = self.parent_plugin_type()
        mapping_parent = type_mapping.get('_parent', None)
        if mapping_parent:
            if mapping_parent['type'] != expected_parent_type:
                raise exception.IndexingException(
                    _("Mapping for '%(doc_type)s' contains a _parent "
                      "'%(actual)s' that doesn't match '%(expected)s'") %
                    {"doc_type": self.document_type,
                     "actual": mapping_parent['type'],
                     "expected": expected_parent_type})
        elif expected_parent_type:
            type_mapping['_parent'] = {'type': expected_parent_type}

        yield self.document_type, type_mapping

    @classmethod
    def get_plugin_type(cls):
        return "resource_plugin"

    @classmethod
    def get_plugin_name(cls):
        return cls.get_document_type().replace("::", "_").lower()

    @property
    def admin_only_fields(self):
        admin_only = self.options.admin_only_fields
        if not admin_only:
            return []
        return self.options.admin_only_fields.split(',')

    @property
    def requires_role_separation(self):
        return len(self.admin_only_fields) > 0

    @classmethod
    def is_plugin_enabled_by_default(cls):
        '''
        Each plugin can overwrite the default value of whether a
        plugin should be enabled if the value is not explicitly
        set in the configuration

        '''
        return True

    @classmethod
    def get_plugin_opts(cls):
        """Options that can be overridden per plugin.
        """
        opts = [
            cfg.StrOpt("resource_group_name"),
            cfg.BoolOpt("enabled", default=cls.is_plugin_enabled_by_default()),
            cfg.StrOpt("admin_only_fields"),
            cfg.BoolOpt('mapping_use_doc_values'),
            cfg.ListOpt('override_region_name',
                        help="Override the region name configured in "
                             "'service_credentials'. This is useful when a "
                             "service is deployed as a cloud-wide service "
                             "rather than per region (e.g. Region1,Region2)."),
            cfg.ListOpt('publishers',
                        help='Used to configure publishers for the plugin, '
                             'value could be publisher names configured in '
                             'setup.cfg file.'
                        )
        ]
        if cls.NotificationHandlerCls:
            opts.extend(cls.NotificationHandlerCls.get_plugin_opts())
        return opts

    @classmethod
    def get_config_group_name(cls):
        """Override the get_plugin_name in order to use the document type as
        plugin name. This turns OS::Service::Resource to os_service_resource
        """
        config_name = cls.get_document_type().replace("::", "_").lower()
        return "resource_plugin:%s" % config_name


@six.add_metaclass(abc.ABCMeta)
class NotificationBase(object):

    def __init__(self, index_helper, options):
        self.index_helper = index_helper
        self.plugin_options = options

    def get_notification_supported_events(self):
        """Get the list of event types this plugin responds to."""
        return list(six.iterkeys(self.get_event_handlers()))

    def get_log_fields(self, event_type, payload):
        """Return an iterable of key value pairs in payload that will be
        present in notification log messages. Document type, event type, timing
        and tenant information will be prepended.
        """
        if 'id' in payload:
            return ('id', payload['id']),
        return ()

    @abc.abstractmethod
    def get_event_handlers(self):
        """Returns a mapping of event name to function"""

    @classmethod
    def _get_notification_exchanges(cls):
        """Return a list of oslo exchanges this plugin cares about"""

    @classmethod
    def get_plugin_opts(cls):
        opts = []
        exchanges = cls._get_notification_exchanges()
        if exchanges:
            defaults = " ".join("<notifications_topic>,%s" % i
                                for i in exchanges)
            opts.append(cfg.MultiOpt(
                'notifications_topics_exchanges',
                item_type=types.MultiString(),
                help='Override default topic,exchange pairs. '
                     'Defaults to %s' % defaults,
                default=[]))
        return opts

    def process(self, ctxt, publisher_id, event_type, payload, metadata):
        """Process the incoming notification message."""
        try:
            docs = self.get_event_handlers()[event_type](
                event_type,
                payload,
                metadata['timestamp'])
            if docs:
                if not hasattr(docs, '__iter__'):
                    docs = [docs]
            return docs
        except Exception as e:
            LOG.error(encodeutils.exception_to_unicode(e))

    def get_notification_topics_exchanges(self):
        """"Returns a list of (topic,exchange), (topic,exchange)..)
        This will either be (CONF.notifications_topic,<exchange>) for
        each exchange in _get_notification_exchanges OR the list of
        values for CONF.plugin.notifications_topics_exchanges.
        """
        configured = self.plugin_options.notifications_topics_exchanges
        if configured:
            return [tuple(i.split(',')) for i in configured]
        else:
            return [(CONF.resource_plugin.notifications_topic, exchange)
                    for exchange in self._get_notification_exchanges()]

    @classmethod
    def get_version(cls, payload, timestamp=None, preferred_date_field=None):
        """Combine <preferred_date_field>|updated_at|created_at epoch with
        notification timestamp as a version number, format is
        <right 9 digits update epoch(create epoch)><right 9 digits timestamp
        in milliseconds>, if timestamp is not present(sync indexing), fill in
        9 digits of zero instead.

        The reason we combine update time and timestamp together is the
        precision of update time is limit to seconds. It's not accurate enough
        to use as a version.

        The max version in Elasticsearch is 9.2e+18, allowing 19 digits at
        most. Our version is 18 digits long, leaves 1 digit for conservation.

        The updated epoch is 10 digits long, we strip off its leading digit,
        concatenate it with the right 9 digits of timestamp in milliseconds,
        and we get a 18 digits long version.

        The truncating has some potential things to be noted, similar to Y2K
        problem.

        Let's say we have an updated epoch 1450161655. Stripping off the
        leading '1' from the current epoch seconds 'rebases' our epoch around
        1984(450161655). By the time we get to an updated epoch beginning '199'
        we're somewhere around 2033, and truncating epoch to 2001. Once the
        clock flips round to begin '200'(year 2033) things will stop working
        because we'll suddenly be using epoch that look like they're from 1969.
        We can address this before that happens; worst case is that you'd have
        to reindex everything, or reset the version.

        The "timestamp" has similar issues. When the "timestamp" overflowed
        the 9-digit field, time becomes indistinguishable. The 9 digits
        millisecond precision gives us around 27 hours. It should be enough to
        distinguish notifications with different timestamps.
        """
        updated = None

        # Pick the best/preferred date field to calculate version from
        date_fields = ('updated_at', 'created_at', 'deleted_at')
        if preferred_date_field:
            date_fields = (preferred_date_field,) + date_fields

        for date_field in date_fields:
            if date_field and payload.get(date_field):
                updated = payload.get(date_field)
                break
        else:
            date_fields_str = ', '.join(date_fields)
            msg = ('Failed to build elasticsearch version; none of %(dfs)s '
                   'found in payload: %(payload)s' %
                   {'dfs': date_fields_str, 'payload': payload})
            raise exception.SearchlightException(message=msg)

        updated_obj = timeutils.parse_isotime(updated)
        updated_epoch = int(calendar.timegm(updated_obj.utctimetuple()))
        if timestamp:
            timestamp_obj = timeutils.parse_isotime(timestamp)
            timestamp_epoch = int(calendar.timegm(
                timestamp_obj.utctimetuple()))
            timestamp_milli = (timestamp_epoch * 1000 +
                               timestamp_obj.microsecond // 1000)
            truncate_timestamp = str(timestamp_milli)[-9:].zfill(9)
            # truncate updated epoch because we are run out of numbers.
            final_stamp = int(str(updated_epoch)[-9:] +
                              str(truncate_timestamp))
        else:
            final_stamp = int(str(updated_epoch)[-9:] + '0' * 9)
        return final_stamp



# Copyright 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools
import operator

from keystoneauth1 import loading

import searchlight.api.versions
import searchlight.common.config
import searchlight.common.property_utils
import searchlight.common.wsgi
import searchlight.elasticsearch
import searchlight.elasticsearch.plugins.base
import searchlight.elasticsearch.plugins.openstack_clients
import searchlight.listener


def list_opts():
    return [
        ('DEFAULT',
         itertools.chain(searchlight.common.property_utils.property_opts,
                         searchlight.common.config.common_opts)),
        ('elasticsearch', searchlight.elasticsearch.search_opts),
        ('service_credentials',
         itertools.chain(
             searchlight.elasticsearch.plugins.openstack_clients.client_opts,
             loading.get_auth_common_conf_options(),
             list_auth_opts())),
        ('resource_plugin',
         searchlight.elasticsearch.plugins.base.indexer_opts),
        ('paste_deploy', searchlight.common.config.paste_deploy_opts),
        ('profiler', searchlight.common.wsgi.profiler_opts),
        ('listener', searchlight.listener.listener_opts),
        ('api',
         itertools.chain(searchlight.api.versions.versions_opts,
                         searchlight.common.wsgi.bind_opts,
                         searchlight.common.wsgi.socket_opts,
                         searchlight.common.wsgi.eventlet_opts)),
    ]


def list_auth_opts():
    # Inspired by similar code in neutron
    opt_list = []
    for plugin in ['password', 'v2password', 'v3password']:
        plugin_options = loading.get_auth_plugin_conf_options(plugin)
        for plugin_option in plugin_options:
            if all(option.name != plugin_option.name for option in opt_list):
                opt_list.append(plugin_option)

    opt_list.sort(key=operator.attrgetter('name'))
    return opt_list



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins import swift
from searchlight.elasticsearch.plugins.swift \
    import serialize_swift_account_notification
from searchlight.elasticsearch.plugins.swift \
    import serialize_swift_container_notification
from searchlight.elasticsearch.plugins.swift \
    import serialize_swift_object_notification
from searchlight import pipeline


LOG = logging.getLogger(__name__)


class SwiftAccountHandler(base.NotificationBase):

    def __init__(self, *args, **kwargs):
        super(SwiftAccountHandler, self).__init__(*args, **kwargs)

    @classmethod
    def _get_notification_exchanges(cls):
        return ['swift']

    def get_event_handlers(self):
        return {
            "account.create": self.create_or_update,
            "account.metadata": self.create_or_update,
            "account.delete": self.delete
        }

    def get_log_fields(self, event_type, payload):
        return ('account', payload.get('account')),

    def create_or_update(self, event_type, payload, timestamp):
        serialized_account = serialize_swift_account_notification(payload)
        try:
            self.index_helper.save_document(
                serialized_account,
                version=self.get_version(serialized_account, timestamp))
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      serialized_account
                                      )
        except Exception as exc:
            LOG.error('Error saving account %(id)s '
                      'in index. Error: %(exc)s' %
                      {'id': payload['id'], 'exc': exc})

    def delete(self, event_type, payload, timestamp):
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        id = payload['account']
        try:
            self.index_helper.delete_document(
                {'_id': id, '_version': version,
                 '_routing': payload['account']})
            return pipeline.DeleteItem(self.index_helper.plugin,
                                       event_type,
                                       payload,
                                       id)
        except Exception as exc:
            LOG.error('Error deleting account %(id)s '
                      'from index. Error: %(exc)s' %
                      {'id': id, 'exc': exc})


class SwiftContainerHandler(base.NotificationBase):

    def __init__(self, *args, **kwargs):
        self.object_helper = kwargs.pop('object_helper')
        super(SwiftContainerHandler, self).__init__(*args, **kwargs)

    @classmethod
    def _get_notification_exchanges(cls):
        return ['swift']

    def get_event_handlers(self):
        return {
            "container.create": self.create_or_update,
            "container.metadata": self.create_or_update,
            "container.delete": self.delete
        }

    def get_log_fields(self, event_type, payload):
        return (
            ('account', payload.get('account')),
            ('container', payload.get('container'))
        )

    def create_or_update(self, event_type, payload, timestamp):
        serialized_payload = serialize_swift_container_notification(payload)
        try:
            self.index_helper.save_document(
                serialized_payload,
                version=self.get_version(serialized_payload, timestamp))
            return pipeline.IndexItem(self.index_helper.plugin,
                                      event_type,
                                      payload,
                                      serialized_payload)
        except Exception as exc:
            LOG.error('Error saving container %(id)s '
                      'in index. Error: %(exc)s' %
                      {'id': payload['id'], 'exc': exc})

    def delete(self, event_type, payload, timestamp):
        # notification payload doesn't have any date/time fields
        # so temporarily use metadata timestamp value as
        # updated_at field to retrieve version
        # Remove it when notification starts sending datetime field(s)
        payload['updated_at'] = timestamp
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        del payload['updated_at']

        id = payload['account'] + swift.ID_SEP + payload['container']
        items = []
        try:
            delete_docs = self.object_helper.delete_documents_with_parent(
                id, version=version)
            items.extend(
                [pipeline.DeleteItem(
                    self.index_helper.plugin,
                    event_type,
                    payload,
                    doc['_id']) for doc in delete_docs])
        except Exception as exc:
            LOG.error('Error deleting objects in container %(id)s '
                      'from index. Error: %(exc)s' %
                      {'id': id, 'exc': exc})
        try:
            self.index_helper.delete_document(
                {'_id': id, '_version': version,
                 '_routing': payload['account']})
            items.append(
                pipeline.DeleteItem(
                    self.index_helper.plugin,
                    event_type,
                    payload,
                )
            )
            return items
        except Exception as exc:
            LOG.error('Error deleting container %(id)s '
                      'from index. Error: %(exc)s' %
                      {'id': id, 'exc': exc})


class SwiftObjectHandler(base.NotificationBase):

    def __init__(self, *args, **kwargs):
        super(SwiftObjectHandler, self).__init__(*args, **kwargs)

    @classmethod
    def _get_notification_exchanges(cls):
        return ['swift']

    def get_event_handlers(self):
        return {
            "object.create": self.create_or_update,
            "object.metadata": self.create_or_update,
            "object.delete": self.delete
        }

    def get_log_fields(self, event_type, payload):
        return (
            ('account', payload.get('account')),
            ('container', payload.get('container')),
            ('object', payload.get('object'))
        )

    def create_or_update(self, event_type, payload, timestamp):
        serialized_payload = serialize_swift_object_notification(payload)
        try:
            self.index_helper.save_document(
                serialized_payload,
                version=self.get_version(serialized_payload, timestamp)
            )
            return pipeline.IndexItem(
                self.index_helper.plugin,
                event_type,
                payload,
                serialized_payload
            )
        except Exception as exc:
            LOG.error('Error saving object %(id)s '
                      'in index. Error: %(exc)s' %
                      {'id': payload['id'], 'exc': exc})

    def delete(self, event_type, payload, timestamp):
        # notification payload doesn't have any date/time fields
        # so temporarily use metadata timestamp value as
        # updated_at field to retrieve version
        # Remove it when notification starts sending datetime field(s)
        payload['updated_at'] = timestamp
        version = self.get_version(payload, timestamp,
                                   preferred_date_field='deleted_at')
        del payload['updated_at']

        id = payload['account'] + swift.ID_SEP + \
            payload['container'] + swift.ID_SEP + \
            payload['object']
        try:
            self.index_helper.delete_document(
                {'_id': id, '_version': version,
                 '_routing': payload['account']})
            return pipeline.DeleteItem(
                self.index_helper.plugin,
                event_type,
                payload,
                id
            )
        except Exception as exc:
            LOG.error('Error deleting object %(id)s '
                      'from index. Error: %(exc)s' %
                      {'id': id, 'exc': exc})



#!/usr/bin/env python
#
# Copyright 2012-2014 eNovance <licensing@enovance.com>
# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys

from oslo_config import cfg
import oslo_i18n
from oslo_log import log
import oslo_messaging
from oslo_reports import guru_meditation_report as gmr

from searchlight.common import utils
from searchlight import version

CONF = cfg.CONF

LOG = log.getLogger(__name__)
_DEFAULT_LOG_LEVELS = ['keystonemiddleware=WARN', 'stevedore=WARN']


class WorkerException(Exception):
    """Exception for errors relating to service workers."""


def prepare_service(argv=None):
    oslo_i18n.enable_lazy()
    log.set_defaults(_DEFAULT_LOG_LEVELS)
    log.register_options(CONF)
    gmr.TextGuruMeditation.setup_autorun(version)
    utils.register_plugin_opts()

    if argv is None:
        argv = sys.argv
    CONF(argv[1:], project='searchlight')
    log.setup(cfg.CONF, 'searchlight')
    oslo_messaging.set_transport_defaults('searchlight')



# Copyright 2015 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import datetime
import mock

import glanceclient.exc

from searchlight.common import utils
from searchlight.elasticsearch.plugins.base import NotificationBase
from searchlight.elasticsearch.plugins.glance import images as images_plugin
from searchlight.elasticsearch import ROLE_USER_FIELD
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


DATETIME = datetime.datetime(2012, 5, 16, 15, 27, 36, 325355)
DATE1 = utils.isotime(DATETIME)

# General
USER1 = '54492ba0-f4df-4e4e-be62-27f4d76b29cf'

TENANT1 = '6838eb7b-6ded-434a-882c-b344c77fe8df'
TENANT2 = '2c014f32-55eb-467d-8fcb-4bd706012f81'
TENANT3 = '5a3e60e8-cfa9-4a9e-a90a-62b42cea92b8'
TENANT4 = 'c6c87f25-8a94-47ed-8c83-053c25f42df4'

# Images
UUID1 = 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d'
UUID2 = 'a85abd86-55b3-4d5b-b0b4-5d0a6e6042fc'
UUID3 = '971ec09a-8067-4bc8-a91f-ae3557f1c4c7'
UUID4 = '6bbe7cc2-eae7-4c0f-b50d-a7160b0c6a86'
UUID5 = 'KERNEL-eae7-4c0f-b50d-RAMDISK'
UUID6 = 'c69d23df-4b3e-4e61-893a-a1dd12200bd3'

CHECKSUM = '93264c3edf5972c9f1cb309543d38a5c'


def _image_fixture(image_id, **kwargs):
    """Simulates a v2 image (which is just a dictionary)
    """
    extra_properties = kwargs.pop('extra_properties', {})

    image = {
        'id': image_id,
        'name': None,
        'visibility': 'public',
        'kernel_id': None,
        'file': 'v2/images/' + image_id,
        'checksum': None,
        'owner': None,
        'status': 'queued',
        'tags': [],
        'size': None,
        'virtual_size': None,
        'locations': [],
        'protected': False,
        'disk_format': None,
        'container_format': None,
        'min_ram': None,
        'min_disk': None,
        'created_at': DATE1,
        'updated_at': DATE1,
    }
    image.update(kwargs)
    for k, v in extra_properties.items():
        image[k] = v
    return image


def _notification_fixture(image_id, **kwargs):
    properties = kwargs.pop('properties', {})
    notification = {
        'id': image_id,
        'name': None,
        'status': 'active',
        'virtual_size': None,
        'deleted': False,
        'disk_format': None,
        'container_format': None,
        'min_ram': None,
        'min_disk': None,
        'protected': False,
        'checksum': None,
        'owner': None,
        'is_public': True,
        'deleted_at': None,
        'size': None,
        'created_at': DATE1,
        'updated_at': DATE1,
        'properties': {},
        'visibility': None
    }
    for k, v in kwargs.items():
        if k in notification:
            notification[k] = v
    for k, v in properties.items():
        notification['properties'][k] = v
    return notification


class TestImageLoaderPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImageLoaderPlugin, self).setUp()
        self.set_property_protections()

        self._create_images()

        self.plugin = images_plugin.ImageIndex()
        self.notification_handler = self.plugin.get_notification_handler()

    def _create_images(self):
        self.simple_image = _image_fixture(
            UUID1, owner=TENANT1, checksum=CHECKSUM, name='simple', size=256,
            visibility='public', status='active'
        )
        self.snapshot_image = _image_fixture(
            UUID1, owner=TENANT1, checksum=CHECKSUM, name='snapshot1',
            size=256, visibility='public', status='active',
            extra_properties={
                'image_type': 'snapshot',
                'base_image_ref': UUID2,
                "user_id": UUID3,
                "instance_uuid": UUID4,
                "image_state": "available"
            }
        )
        self.tagged_image = _image_fixture(
            UUID2, owner=TENANT1, checksum=CHECKSUM, name='tagged', size=512,
            visibility='public', status='active', tags=['ping', 'pong'],
        )
        self.complex_image = _image_fixture(
            UUID3, owner=TENANT2, checksum=CHECKSUM, name='complex', size=256,
            visibility='public', status='active',
            extra_properties={'mysql_version': '5.6', 'hypervisor': 'lxc'}
        )
        self.members_image = _image_fixture(
            UUID6, owner=TENANT2, checksum=CHECKSUM, name='complex', size=256,
            visibility='shared', status='active',
        )
        self.private_image = _image_fixture(
            UUID3, owner=TENANT2, checksum=CHECKSUM, name='complex', size=256,
            visibility='private', status='active',
        )
        self.members_image_members = [
            {'member_id': TENANT1, 'status': 'accepted'},
            {'member_id': TENANT2, 'status': 'accepted'},
            {'member_id': TENANT3, 'status': 'accepted'},
            {'member_id': TENANT4, 'status': 'pending'},
        ]
        self.kernel_ramdisk_image = _image_fixture(
            UUID5, owner=TENANT1, checksum=CHECKSUM, name='kernel_ramdisk',
            size=256, visibility='public', status='active',
            kernel_id='KERNEL-ID-SEARCH-LIGHT-ROCKS',
            ramdisk_id='RAMDISK-ID-GO-BRONCOS',
        )
        self.images = [self.simple_image, self.tagged_image,
                       self.complex_image, self.members_image,
                       self.kernel_ramdisk_image, self.private_image]

    def test_resource_group_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Glance::Image', self.plugin.get_document_type())

    def test_image_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d',
            'image_type': 'image',
            'kernel_id': None,
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'simple',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1
        }
        serialized = self.plugin.serialize(self.simple_image)
        self.assertEqual(expected, serialized)

    def test_image_snapshot_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': UUID1,
            'image_type': 'image',
            'kernel_id': None,
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'snapshot1',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1,
            "image_type": "snapshot",
            "base_image_ref": UUID2,
            "user_id": UUID3,
            "instance_uuid": UUID4,
            "image_state": "available",
        }
        serialized = self.plugin.serialize(self.snapshot_image)
        self.assertEqual(expected, serialized)

    def test_image_with_tags_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': 'a85abd86-55b3-4d5b-b0b4-5d0a6e6042fc',
            'image_type': 'image',
            'kernel_id': None,
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'tagged',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 512,
            'status': 'active',
            'tags': ['ping', 'pong'],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1
        }
        serialized = self.plugin.serialize(self.tagged_image)
        self.assertEqual(expected, serialized)

    def test_image_with_properties_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'hypervisor': 'lxc',
            'id': '971ec09a-8067-4bc8-a91f-ae3557f1c4c7',
            'image_type': 'image',
            'kernel_id': None,
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'mysql_version': '5.6',
            'name': 'complex',
            'owner': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'project_id': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1
        }

        serialized = self.plugin.serialize(self.complex_image)
        self.assertEqual(expected, serialized)

    def test_image_with_members_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': 'c69d23df-4b3e-4e61-893a-a1dd12200bd3',
            'image_type': 'image',
            'kernel_id': None,
            'members': ['6838eb7b-6ded-434a-882c-b344c77fe8df',
                        '2c014f32-55eb-467d-8fcb-4bd706012f81',
                        '5a3e60e8-cfa9-4a9e-a90a-62b42cea92b8'],
            'min_disk': None,
            'min_ram': None,
            'name': 'complex',
            'owner': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'project_id': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'shared',
            'created_at': DATE1,
            'updated_at': DATE1
        }

        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        return_value=self.members_image_members):
            serialized = self.plugin.serialize(self.members_image)
        self.assertEqual(expected, serialized)

    def test_image_kernel_ramdisk_serialize(self):
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': UUID5,
            'image_type': 'image',
            'kernel_id': 'KERNEL-ID-SEARCH-LIGHT-ROCKS',
            'ramdisk_id': 'RAMDISK-ID-GO-BRONCOS',
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'kernel_ramdisk',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1
        }
        serialized = self.plugin.serialize(self.kernel_ramdisk_image)
        self.assertEqual(expected, serialized)

    def test_index_initial_data(self):
        """Tests initial data load."""
        image_member_mocks = [
            self.members_image_members
        ]
        with mock.patch('glanceclient.v2.images.Controller.list',
                        return_value=self.images) as mock_list:
            with mock.patch('glanceclient.v2.image_members.Controller.list',
                            side_effect=image_member_mocks) as mock_members:
                # This is not testing the elasticsearch call, just
                # that the documents being indexed are as expected
                with mock.patch.object(
                        self.plugin.index_helper,
                        'save_documents') as mock_save:
                    self.plugin.index_initial_data()
                    versions = [NotificationBase.get_version(img)
                                for img in self.images]
                    mock_list.assert_called_once_with()
                    mock_members.assert_called_once_with(
                        self.members_image['id'])

                    mock_save.assert_called_once_with([
                        {
                            'kernel_id': None,
                            'tags': [],
                            'protected': False,
                            'min_disk': None,
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 256,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [],
                            'visibility': 'public',
                            'owner': TENANT1,
                            'project_id': TENANT1,
                            'disk_format': None,
                            'name': 'simple',
                            'created_at': '2012-05-16T15:27:36Z',
                            'id': 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d',
                            'image_type': 'image'
                        },
                        {
                            'kernel_id': None,
                            'tags': ['ping', 'pong'],
                            'protected': False,
                            'min_disk': None,
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 512,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [],
                            'visibility': 'public',
                            'owner': TENANT1,
                            'project_id': TENANT1,
                            'disk_format': None,
                            'name': 'tagged',
                            'created_at': '2012-05-16T15:27:36Z',
                            'id': 'a85abd86-55b3-4d5b-b0b4-5d0a6e6042fc',
                            'image_type': 'image'
                        },
                        {
                            'kernel_id': None,
                            'tags': [], 'protected': False,
                            'min_disk': None,
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 256,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [],
                            'mysql_version': '5.6',
                            'visibility': 'public',
                            'hypervisor': 'lxc',
                            'owner': TENANT2,
                            'project_id': TENANT2,
                            'id': '971ec09a-8067-4bc8-a91f-ae3557f1c4c7',
                            'image_type': 'image',
                            'name': 'complex',
                            'created_at': '2012-05-16T15:27:36Z',
                            'disk_format': None
                        },
                        {
                            'kernel_id': None, 'tags': [],
                            'protected': False,
                            'min_disk': None,
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 256,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [
                                '6838eb7b-6ded-434a-882c-b344c77fe8df',
                                '2c014f32-55eb-467d-8fcb-4bd706012f81',
                                '5a3e60e8-cfa9-4a9e-a90a-62b42cea92b8'
                            ],
                            'visibility': 'shared',
                            'owner': TENANT2,
                            'project_id': TENANT2,
                            'disk_format': None,
                            'name': 'complex',
                            'created_at': '2012-05-16T15:27:36Z',
                            'id': 'c69d23df-4b3e-4e61-893a-a1dd12200bd3',
                            'image_type': 'image'
                        },
                        {
                            'kernel_id': 'KERNEL-ID-SEARCH-LIGHT-ROCKS',
                            'tags': [], 'protected': False,
                            'min_disk': None,
                            'ramdisk_id': 'RAMDISK-ID-GO-BRONCOS',
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 256,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [],
                            'visibility': 'public',
                            'owner': TENANT1,
                            'project_id': TENANT1,
                            'disk_format': None,
                            'name': 'kernel_ramdisk',
                            'created_at': '2012-05-16T15:27:36Z',
                            'id': 'KERNEL-eae7-4c0f-b50d-RAMDISK',
                            'image_type': 'image'
                        },
                        {
                            'kernel_id': None, 'tags': [],
                            'protected': False,
                            'min_disk': None,
                            'min_ram': None,
                            'virtual_size': None,
                            'size': 256,
                            'container_format': None,
                            'status': 'active',
                            'updated_at': '2012-05-16T15:27:36Z',
                            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
                            'members': [],
                            'visibility': 'private',
                            'owner': TENANT2,
                            'project_id': TENANT2,
                            'disk_format': None,
                            'name': 'complex',
                            'created_at': '2012-05-16T15:27:36Z',
                            'id': '971ec09a-8067-4bc8-a91f-ae3557f1c4c7',
                            'image_type': 'image'
                        }
                    ], index=None, versions=versions)

    def test_image_rbac(self):
        """Test the image plugin RBAC query terms"""
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search'
        )
        rbac_query_fragment = self.plugin.get_query_filters(
            fake_request.context)
        expected_fragment = {
            "indices": {
                "index": "searchlight-search",
                "no_match_query": "none",
                "query": {
                    "bool": {
                        "filter": {
                            "bool": {
                                "must": {
                                    "type": {"value": "OS::Glance::Image"}
                                },
                                "should": [
                                    {"term": {"owner": TENANT1}},
                                    {"term": {"visibility": "public"}},
                                    {"term": {"members": TENANT1}}
                                ],
                                'minimum_should_match': 1
                            }
                        }
                    }
                },
            }
        }

        self.assertEqual(expected_fragment, rbac_query_fragment)

    def test_protected_properties(self):
        extra_props = {
            'x_foo_matcher': 'this is protected',
            'x_foo_something_else': 'this is not protected',
            'z_this_has_no_rules': 'this is protected too'
        }
        image_with_properties = _image_fixture(
            UUID1, owner=TENANT1, checksum=CHECKSUM, name='simple', size=256,
            status='active', extra_properties=extra_props
        )

        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        return_value=[]):
            serialized = self.plugin.serialize(image_with_properties)

        elasticsearch_results = {
            'hits': {
                'hits': [{
                    '_source': copy.deepcopy(serialized),
                    '_type': self.plugin.get_document_type(),
                    '_index': self.plugin.alias_name_search
                }]
            }
        }

        # Admin context
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=True
        )

        for result_hit in elasticsearch_results['hits']['hits']:
            self.plugin.filter_result(result_hit, fake_request.context)

        # This should contain the three properties we added
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d',
            'image_type': 'image',
            'kernel_id': None,
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'simple',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1,
            'x_foo_matcher': 'this is protected',
            'x_foo_something_else': 'this is not protected',
            'z_this_has_no_rules': 'this is protected too'
        }

        self.assertEqual(expected,
                         elasticsearch_results['hits']['hits'][0]['_source'])

        # Non admin user. Recreate this because the filter operation modifies
        # it in place and we want a fresh copy
        elasticsearch_results = {
            'hits': {
                'hits': [{
                    '_source': copy.deepcopy(serialized),
                    '_type': self.plugin.get_document_type(),
                    '_index': self.plugin.alias_name_search
                }]
            }
        }
        # Non admin context should miss the x_foo property
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )

        for result_hit in elasticsearch_results['hits']['hits']:
            self.plugin.filter_result(result_hit, fake_request.context)

        # Should be missing two of the properties
        expected = {
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'container_format': None,
            'disk_format': None,
            'id': 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d',
            'members': [],
            'min_disk': None,
            'min_ram': None,
            'name': 'simple',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'protected': False,
            'size': 256,
            'status': 'active',
            'tags': [],
            'virtual_size': None,
            'visibility': 'public',
            'created_at': DATE1,
            'updated_at': DATE1,
            'x_foo_something_else': 'this is not protected'
        }

        self.assertEqual(expected,
                         elasticsearch_results['hits']['hits'][0]['_source'])

    def test_image_notification_serialize(self):
        notification = _notification_fixture(
            self.simple_image['id'],
            checksum=self.simple_image['checksum'],
            name=self.simple_image['name'],
            is_public=True,
            size=self.simple_image['size'],
            properties={'prop1': 'val1'},
            owner=self.simple_image['owner'])

        expected = {
            'status': 'active',
            # Tags are not contained in notifications
            # 'tags': [],
            'container_format': None,
            'min_ram': None,
            'visibility': 'public',
            'owner': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'project_id': '6838eb7b-6ded-434a-882c-b344c77fe8df',
            'min_disk': None,
            'members': [],
            'virtual_size': None,
            'id': 'c80a1a6c-bd1f-41c5-90ee-81afedb1d58d',
            'image_type': 'image',
            'size': 256,
            'prop1': 'val1',
            'name': 'simple',
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'disk_format': None,
            'protected': False,
            'created_at': DATE1,
            'updated_at': DATE1
        }

        serialized = self.notification_handler.serialize_notification(
            notification)
        self.assertEqual(expected, serialized)

    def test_private_image_notification_serialize(self):
        """Test a notification for a private image"""
        notification = _notification_fixture(
            self.members_image['id'],
            checksum=self.members_image['checksum'],
            name=self.members_image['name'],
            is_public=False,
            size=self.members_image['size'],
            owner=self.members_image['owner'])

        expected = {
            'status': 'active',
            # Tags are not contained in notifications
            # 'tags': [],
            'container_format': None,
            'min_ram': None,
            'visibility': 'private',
            'owner': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'project_id': '2c014f32-55eb-467d-8fcb-4bd706012f81',
            'members': [],
            'min_disk': None,
            'virtual_size': None,
            'id': 'c69d23df-4b3e-4e61-893a-a1dd12200bd3',
            'image_type': 'image',
            'size': 256,
            'name': 'complex',
            'checksum': '93264c3edf5972c9f1cb309543d38a5c',
            'disk_format': None,
            'protected': False,
            'created_at': DATE1,
            'updated_at': DATE1,
        }
        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        return_value=self.members_image_members):
            serialized = self.notification_handler.serialize_notification(
                notification)
        self.assertEqual(expected, serialized)

    def test_facets(self):
        """Check that expected fields are faceted"""
        mock_engine = mock.Mock()
        self.plugin.engine = mock_engine
        mock_engine.search.return_value = {'aggregations': {},
                                           'hits': {'total': 0}}

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search/facets', is_admin=False
        )

        facets, _ = self.plugin.get_facets(fake_request.context)
        facet_names = [f['name'] for f in facets]

        # Glance has no nested fields but owner and project id are excluded for
        # non admins
        expected_facet_names = self.plugin.get_mapping()['properties'].keys()
        expected_facet_names = set(expected_facet_names) - set(('owner',
                                                                'project_id'))

        self.assertEqual(set(expected_facet_names), set(facet_names))

        facet_option_fields = ('disk_format', 'container_format', 'tags',
                               'visibility', 'status', 'protected',
                               'image_type', 'image_state')
        expected_agg_query = {
            'aggs': dict(unit_test_utils.simple_facet_field_agg(name)
                         for name in facet_option_fields),
            'query': {
                'bool': {
                    'filter': {
                        'bool': {
                            'must': {'term': {ROLE_USER_FIELD: 'user'}},
                            'should': [
                                {'term': {'owner': TENANT1}},
                                {'term': {'visibility': 'public'}},
                                {'term': {'members': TENANT1}}
                            ],
                            'minimum_should_match': 1
                        }
                    }
                }
            }
        }

        mock_engine.search.assert_called_with(
            index=self.plugin.alias_name_search,
            doc_type=self.plugin.get_document_type(),
            body=expected_agg_query,
            ignore_unavailable=True,
            size=0
        )

    @mock.patch('searchlight.elasticsearch.plugins.helper.IndexingHelper.'
                'delete_document')
    def test_create_or_update_exception(self, mock_delete):
        notification = _notification_fixture(
            self.members_image['id'],
            checksum=self.members_image['checksum'],
            name=self.members_image['name'],
            is_public=False,
            size=self.members_image['size'],
            owner=self.members_image['owner'],
            visibility='shared')

        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        side_effect=glanceclient.exc.NotFound):
            with mock.patch.object(
                    self.notification_handler,
                    'get_version',
                    return_value='fake_version'):
                self.notification_handler.create_or_update(
                    "image.create", notification, None)
                mock_delete.assert_called_with(
                    {'_id': self.members_image['id'],
                     '_version': 'fake_version'})

    def test_image_member_list(self):
        with mock.patch('glanceclient.v2.image_members.Controller.list',
                        return_value=self.members_image_members):
            serialized = self.plugin.serialize(self.members_image)

        # Admin context
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=True
        )
        result_hit = {'_source': copy.deepcopy(serialized)}

        # Expect to see all three tenants
        self.plugin.filter_result(result_hit, fake_request.context)
        self.assertEqual(set([TENANT1, TENANT2, TENANT3]),
                         set(result_hit['_source']['members']))

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )
        result_hit = {'_source': copy.deepcopy(serialized)}

        # Tenant1 can see the image but doesn't own it
        self.plugin.filter_result(result_hit, fake_request.context)
        self.assertEqual(set([TENANT1]),
                         set(result_hit['_source']['members']))

        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT2, '/v1/search', is_admin=False
        )
        result_hit = {'_source': copy.deepcopy(serialized)}

        # Tenant2 owns the image and should see all three members
        self.plugin.filter_result(result_hit, fake_request.context)
        self.assertEqual(set([TENANT1, TENANT2, TENANT3]),
                         set(result_hit['_source']['members']))


class TestImageEventHandler(test_utils.BaseTestCase):
    def setUp(self):
        super(TestImageEventHandler, self).setUp()
        self.set_property_protections()

        self._create_images()

        self.plugin = images_plugin.ImageIndex()
        self.notification_handler = self.plugin.get_notification_handler()



# Copyright (c) 2016 Hewlett-Packard Enterprise Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import time

from searchlight.common import utils
import searchlight.elasticsearch.plugins.swift as swift_plugin
from searchlight.elasticsearch.plugins.swift import\
    accounts as accounts_plugin
import searchlight.tests.unit.utils as unit_test_utils
import searchlight.tests.utils as test_utils


now_epoch_time = time.time()
now_utc = datetime.datetime.fromtimestamp(now_epoch_time)\
    .strftime('%Y-%m-%dT%H:%M:%SZ')


USER1 = u'27f4d76b-be62-4e4e-aa33bb11cc55'
ID1 = "AUTH_488ac936-663e-4e5c-537d-986021b32c4b"
ID2 = "AUTH_7554da43-6443-acdf-deac-3425223cdada"
ID3 = "AUTH_30754354-ca43-124b-12b5-789234bcdefa'"

AUTH_PREFIX = "AUTH_"

X_ACCOUNT_META_KEY1 = 'x-account-meta-key1'
X_ACCOUNT_META_VALUE1 = 'x-account-meta-value1'

X_ACCOUNT_META_KEY2 = 'x-account-meta-key2'
X_ACCOUNT_META_VALUE2 = 'x-account-meta-value2'

TENANT1 = "15b9a454cee34dbe9933ad575a0a6930"
DOMAIN_ID = "default"

DATETIME = datetime.datetime(2016, 2, 20, 1, 13, 24, 215337)
DATE1 = utils.isotime(DATETIME)


def _account_fixture(account_id, domain_id, name, **kwargs):
    fixture = {
        "id": account_id,
        "name": name,
        "x-account-project-domain-id": domain_id,
        'x-timestamp': now_epoch_time
    }
    fixture.update(kwargs)
    return fixture


def _notification_account_fixture(account_id, **kwargs):
    metadata = kwargs.pop('meta', {})
    notification = {
        'account': account_id,
        'project_name': None,
        'updated_at': DATE1,
        'project_domain_name': None,
        'x-trans-id': None,
        'project_id': None,
        'project_domain_id': None
    }
    for k, v in kwargs.items():
        if k in notification:
            notification[k] = v
    for k, v in metadata.items():
        notification[k] = v
    return notification


class TestSwiftAccountPlugin(test_utils.BaseTestCase):
    def setUp(self):
        super(TestSwiftAccountPlugin, self).setUp()
        self.plugin = accounts_plugin.AccountIndex()
        self._create_fixtures()

    def _create_fixtures(self):
        self.account1 = _account_fixture(
            account_id=ID1, domain_id=DOMAIN_ID, name="test-account1")
        self.account2 = _account_fixture(
            account_id=ID2, domain_id=DOMAIN_ID, name="test-account1",
            **{X_ACCOUNT_META_KEY1: X_ACCOUNT_META_VALUE1})
        self.account3 = _account_fixture(
            account_id=ID3, domain_id=DOMAIN_ID, name="test-account1",
            **{X_ACCOUNT_META_KEY1: X_ACCOUNT_META_VALUE1,
                X_ACCOUNT_META_KEY2: X_ACCOUNT_META_VALUE2})
        self.accounts = [self.account1, self.account2, self.account3]

    def test_index_name(self):
        self.assertEqual('searchlight', self.plugin.resource_group_name)

    def test_document_type(self):
        self.assertEqual('OS::Swift::Account',
                         self.plugin.get_document_type())

    def test_rbac_filter(self):
        fake_request = unit_test_utils.get_fake_request(
            USER1, TENANT1, '/v1/search', is_admin=False
        )
        rbac_terms = self.plugin._get_rbac_field_filters(fake_request.context)
        self.assertEqual(
            [{"term": {"id": AUTH_PREFIX + TENANT1}}],
            rbac_terms
        )

    def test_admin_only_fields(self):
        admin_only_fields = self.plugin.admin_only_fields
        self.assertEqual(['x-account-meta-temp-url-key',
                          'x-account-meta-temp-url-key-2',
                          'x-account-access-control'], admin_only_fields)

    def test_serialize(self):
        serialized = self.plugin.serialize(self.account1)

        self.assertEqual(ID1, serialized['id'])
        self.assertEqual('test-account1', serialized['name'])
        self.assertEqual(DOMAIN_ID, serialized['domain_id'])
        self.assertEqual(now_utc, serialized['created_at'])

        serialized = self.plugin.serialize(self.account2)
        self.assertEqual(X_ACCOUNT_META_VALUE1,
                         serialized[X_ACCOUNT_META_KEY1])

        serialized = self.plugin.serialize(self.account3)
        self.assertEqual(X_ACCOUNT_META_VALUE1,
                         serialized[X_ACCOUNT_META_KEY1])
        self.assertEqual(X_ACCOUNT_META_VALUE2,
                         serialized[X_ACCOUNT_META_KEY2])

    def test_swift_account_notification_serialize(self):
        notification = _notification_account_fixture(
            ID1,
            project_name='admin',
            project_id=ID1,
            project_domain_id='default',
            updated_at=DATE1,
        )

        expected = {
            'id': ID1,
            'name': 'admin',
            'updated_at': DATE1,
            'domain_id': 'default'
        }

        serialized = swift_plugin.serialize_swift_account_notification(
            notification)
        self.assertEqual(expected, serialized)



# Copyright (c) 2016 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg

from searchlight.common import resource_types
from searchlight.elasticsearch.plugins import base
from searchlight.elasticsearch.plugins.swift import get_swift_objects
from searchlight.elasticsearch.plugins.swift import serialize_swift_object
from searchlight.elasticsearch.plugins.swift import swift_notification_handler

swift_plugin_opts = [
    cfg.StrOpt('reseller_prefix', default="AUTH_",
               help="prefix used in account names for auth system."),
]

CONF = cfg.CONF
CONF.register_opts(swift_plugin_opts, group='resource_plugin:os_swift_account')


class ObjectIndex(base.IndexBase):
    NotificationHandlerCls = swift_notification_handler.SwiftObjectHandler

    @classmethod
    def parent_plugin_type(cls):
        return resource_types.SWIFT_CONTAINER

    @classmethod
    def get_document_type(cls):
        return resource_types.SWIFT_OBJECT

    def get_mapping(self):
        return {
            'dynamic': True,
            'properties': {
                'id': {'type': 'string', 'index': 'not_analyzed'},
                'name': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    },
                },
                'account': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    },
                },
                'account_id': {'type': 'string', 'index': 'not_analyzed'},
                'container': {
                    'type': 'string',
                    'fields': {
                        'raw': {'type': 'string', 'index': 'not_analyzed'}
                    },
                },
                'container_id': {'type': 'string', 'index': 'not_analyzed'},
                'content_type': {'type': 'string', 'index': 'not_analyzed'},
                'content_length': {'type': 'long'},
                'etag': {'type': 'string'},
                'created_at': {'type': 'date'},
                'updated_at': {'type': 'date'}
            },
            "_parent": {
                "type": self.parent_plugin_type()
            },
            "_meta": {
                "account_id": {
                    "resource_type": resource_types.SWIFT_ACCOUNT
                }
            }
        }

    @property
    def allow_admin_ignore_rbac(self):
        return False

    @classmethod
    def is_plugin_enabled_by_default(cls):
        return False

    @property
    def resource_allowed_policy_target(self):
        return None

    @property
    def service_type(self):
        return None

    def _get_rbac_field_filters(self, request_context):

        tenant_member = request_context.tenant + ":*"
        single_user = request_context.tenant + ":" + request_context.user
        account_plugin = self.parent_plugin.parent_plugin
        account_id = \
            account_plugin.options.reseller_prefix + request_context.owner

        return [
            {
                "has_parent": {
                    "type": self.parent_plugin_type(),
                    "query": {
                        "bool": {
                            "should": [
                                {'term': {'account_id': account_id}},
                                {'terms': {
                                    'x-container-read': [
                                        tenant_member,
                                        single_user
                                    ]
                                }}
                            ]
                        }
                    }
                }
            }
        ]

    def get_parent_id_field(self):
        return 'container_id'

    @property
    def routing_field(self):
        return "account_id"

    @property
    def facets_with_options(self):
        return ('container', 'content_type')

    def get_objects(self):
        return get_swift_objects()

    def serialize(self, obj):
        return serialize_swift_object(obj)
